/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *    (c) Copyright 2001 Phil Schwan <phil@off.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>

#include <readline/readline.h>
#include <readline/history.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include <sys/un.h>
#include <sys/socket.h>

#include <glib.h>

#include "main.h"

/*
 * haha globals.  nya, nya.
 */
static GList *command_list = NULL;
void *shm_buf = NULL;
int shm_id;
int fd;
struct sockaddr_un sun, server_sun;

/*
 * we maintain a list of commands.. some commands have more commands
 * under them as another list.. 
 */
void 
command_register(struct command *cmd)
{
    command_list = g_list_append(command_list, cmd);
}

static void 
list_commands(GList *list, int depth)
{
    do {
        struct command *cmd = list->data;

        printf("%*s", depth, cmd->name);

        if ( cmd->arg_help )
            printf("  %s", cmd->arg_help);

        if ( cmd->help )
            printf("\n%*s  %s", depth, " ", cmd->help);

        printf("\n");

        if ( cmd->children) 
            list_commands(cmd->children, depth + 8);

    } while ( (list = g_list_next(list)) != NULL);
}

/*
 * we have a pretty lame global parser for being able to poke at
 * lines that are issued.. this mostly exists to deal with the
 * amazingly crazy readline tab completion interface
 */
static int cmdc;
static int remc;
#define MAX_CMDC 1024
static char *cmdv[MAX_CMDC];
GList *tree_path[MAX_CMDC];
char *line_copy = NULL;

/*
 * match our token list with a path down the command tree..  cmdc is set to the
 * number of words from the start that matched a path in the command tree..
 * remc are the remaining words.  tree_path records the list item of commands
 * that matched.
 */
static void 
walk_tokens(void) 
{ 
    int i; 
    GList *level = command_list;

    for ( i = 0 ; i < cmdc ; i++ ) {
        char *str = cmdv[i];
        GList *list;

        for ( list = level; list ; list = g_list_next(list)) {
            struct command *cmd = list->data;

            if (g_strcasecmp(cmd->name, str))
                continue;

            tree_path[i] = list;
            level = cmd->children;
            break;
        }

        if ( list == NULL )  {
            break;
        }
    }

    remc = cmdc - i;
    cmdc = i;
}

/*
 * make a copy of a line, tokenize it, and walk the command tree..
 */
static int
parse_line(char *source_line)
{
    char *str, *tok;
    char *line;

    memset(cmdv, 0, sizeof(cmdv));
    cmdc = 0;
    remc = 0;

    if ( line_copy )
        free(line_copy);

    line = line_copy = strdup(source_line);

    for ( str = line; (tok = strsep( &str, " \t")) ; ) {

        if ( tok == NULL || *tok == '\0')
            break;

        g_strstrip(tok);
        cmdv[cmdc++] = tok;

        if ( str == NULL )
            break;
    }

    walk_tokens();

    return cmdc;
}

static void
parse_tokens(int argc, char **argv)
{
    int i;

    memset(cmdv, 0, sizeof(cmdv));
    cmdc = argc;
    remc = 0;

    for(i = 0; i < argc; i++ ) {
        cmdv[i] = argv[i];
        g_strstrip(argv[i]);
    }

    walk_tokens();
}

/*
 * try and execute the path we have through the command tree..
 */
static void
run_tokens(void)
{
    GList *list;
    struct command *cmd;

    if ( cmdc == 0 ) {
        if ( remc ) 
            printf("unknown command '%s'\n", cmdv[0]);
        return;
    }

    list = tree_path[cmdc - 1];
    cmd = list->data;

    /*
     * if the last thing we parsed had children then
     * its an incomplete command
     */
    if ( cmd->children ) {
        list_commands(list, 8);
        return;
    }

    if ( cmd->func == NULL ) {
        printf("oh my, trying to run null command '%s'\n", cmd->name);
        return;
    }

    if (cmd->num_args && cmd->num_args != remc) {
        printf("was expecting %d args.\n", cmd->num_args);
        if ( cmd->arg_help ) 
            printf("\t%s\n", cmd->arg_help);
        return;
    }

    cmd->func(remc, cmdv + cmdc, shm_buf);
}

/*
 * insane readline command completion
 */
GList *gen_list;

/*
 * we're given a list of commands that we should try to match
 * with this partial word.. for now we only match commands, not
 * arguments.
 */
char *
command_generator(const char *text, int state)
{
    int len = strlen(text);

    while(gen_list) {
        struct command *cmd = gen_list->data;

        gen_list = g_list_next(gen_list);

        if ( len == 0 || strncmp(cmd->name, text, len) == 0 ) 
            return strdup(cmd->name);
    }

    return NULL;
}

/*
 * we're trying to tab complete a buffer.  we parse the line
 * to find the last matched command and complete the dangling
 * text at the end against the children of the last completed
 * command..
 */
static char **
cmd_completion(const char *text, int start, int end)
{
    parse_line(rl_line_buffer);

    if ( cmdc ) {
        struct command *cmd = tree_path[cmdc - 1]->data;
        gen_list = cmd->children;
    } else 
        gen_list = command_list;

    return rl_completion_matches (text, command_generator);
}

static void 
run_prompt(void)
{
    char *line;

    rl_attempted_completion_function = cmd_completion;

    for ( ;  (line = readline("cebcfg> ")) != NULL; free(line)) {

        g_strstrip(line);

        if ( *line == '\0')
            continue;

        add_history(line);

        parse_line(line);
        run_tokens();
    }

    if ( line == NULL )
        printf("\n");
}

/*
 * some quick internal commands..
 */

CMD_FUNC(exit_func) {
    exit(1);
}
CMD_FUNC(help_func) {
    list_commands(command_list, 8);
}

static struct command exit_cmd = {
    name:       "exit",
    func:       exit_func,
    children:   NULL,
    help:      "exit",
};
static struct command quit_cmd = {
    name:       "quit",
    func:       exit_func,
    children:   NULL,
    help:      "quit",
};
static struct command help_cmd = {
    name:       "help",
    func:       help_func,
    children:   NULL,
    help:      "list available commands",
};
static struct command q_cmd = {
    name:       "?",
    func:       help_func,
    children:   NULL,
    help:      "list available commands",
};

/*
 * the servers have a unix domain socket sitting open and waiting for
 * commands.  we open a shared mem buf, fill it with command args, and
 * send the command id and shared mem id across the socket.  the server
 * maps in the shared mem region, reads the args, and sends a return
 * code over the socket, filling the shared mem with any response..
 */

static void 
shm_attach(int size)
{
    int reps = 5;

    if ( shm_buf ) {
        shmdt(shm_buf);
        shm_buf = NULL;
    }

    while (reps--) {
        key_t key = rand();

        shm_id = shmget(key, size, IPC_CREAT|IPC_EXCL|0770); 

        if ( shm_id != -1 )
            break;
        else if ( errno != EEXIST )
            break;
    }

    if ( shm_id == -1 ) {
        perror("shmget");
        exit(1);
    }

    shm_buf = shmat(shm_id, 0, 0);
    if ( shm_buf == (void *)-1 ) {
        perror("shmat");
        exit(1);
    }

    shmctl(shm_id, IPC_RMID, NULL);
    memset(shm_buf, 0, size);

}

static void 
unlink_sun(void)
{
    unlink(sun.sun_path);
}

static char socket_path[] = "/tmp/";

static void 
init_socket(char *server_path)
{
    int i;

    fd = socket(AF_LOCAL, SOCK_DGRAM, 0);
    if ( fd < 0 ) {
        perror("socket(AF_LOCAL, SOCK_DGRAM, 0)");
        exit(1);
    }

    sun.sun_family = AF_LOCAL;
    strcpy(sun.sun_path, socket_path);
    strcat(sun.sun_path, "cebcfg_");

    for( i = 0; i < 20 ; i++ ) {
        char tmp[2] = "A";

        tmp[0] = 'A' + (rand() % 26);

        strcat(sun.sun_path, tmp);
    }

    if ( bind(fd, (struct sockaddr *)&sun, sizeof(struct sockaddr_un))) {
        perror("bind()ing to our socket");
        exit(1);
    }

    atexit(unlink_sun);

    server_sun.sun_family = AF_LOCAL;
    strcpy(server_sun.sun_path, server_path);

    if ( connect(fd, (struct sockaddr *)&server_sun, 
            sizeof(struct sockaddr_un))) {
        perror("connect()ing to cebollad");
        exit(1);
    }

}

int 
daemon_command(int cmd)
{
    int ret;
    int args[2] = { cmd, shm_id };

    write(fd, args, sizeof(args));
    read(fd, &ret, sizeof(ret));

    return ret;
};

int main(int argc, char **argv)
{
	int c;
	int prompt = 0;
    int size = 512*1024;
    struct timeval tv;
    char *server_path = "/tmp/cmd_socket";

    command_register(&help_cmd);
    command_register(&q_cmd);
    command_register(&exit_cmd);
    command_register(&quit_cmd);

    gettimeofday(&tv, NULL);

    srand(getpid() ^ tv.tv_sec ^ tv.tv_usec);

    if ( argc == 1 ) {
        printf("%s [ -h -c -p cmd_socket_path -s initial_shm_buf_size ]\n", 
                argv[0]);
        exit(1);
    }

	while ( (c = getopt(argc, argv, "p:s:ch"))  > 0) {
		switch (c) {
        case 'p':
            server_path = strdup(optarg);
            break;
        case 's':
            size = atol(optarg);
            break;
		case 'h':
            list_commands(command_list, 8);
            exit(1);
            break;
		case 'c':
			prompt = 1;
            break;
        }
    }

    shm_attach(size);
    init_socket(server_path);

    if ( prompt )
        run_prompt();

    parse_tokens(argc - optind, argv + optind);
    run_tokens();

    return 0;
}
