/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2001 Phil Schwan <phil@off.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>

#include <glib.h>

#include "auth.h"
#include "log.h"
#include "pkt.h"
#include "debug.h"
#include "util.h"

/*
 * Provide the infrastructure for pluggable auth types.
 */

static GHashTable *auth_hash = NULL;

static void
auth_append_one(gpointer key, gpointer value, gpointer data)
{
    struct auth_type *at = (struct auth_type *)value;
    struct packet *pkt = (struct packet *)data;
    guint16 id;

    if (u16_hash(key) != at->auth_id)
        l_warn("hash key != at->auth_id (%u != %u); trusting auth_id, but "
               "this is bad!\n", u16_hash(key), at->auth_id);

    id = htons(at->auth_id);
    memcpy(pkt_push_tail(pkt, sizeof(id)), &id, sizeof(id));
}

/* append a list of 16-bit auth IDs that we understand */
int
auth_append_list(struct packet *pkt)
{
    g_hash_table_foreach(auth_hash, auth_append_one, (gpointer)pkt);

    return 0;
}
/* 
 * choose an auth method; for now, just pick the first one that we support.  we
 * can put more clever heuristics or configuration options in later.
 */
int
auth_choose(guint16 *list, int num)
{
    while (num--) {
        guint16 local;
        memcpy(&local, &list[num], sizeof(guint16));
        local = ntohs(local);

        if (g_hash_table_lookup(auth_hash, &local) != NULL)
            return local;
    }

    return -1;
}

/* Instantiate a new auth_instance of the specified type. */
struct auth_instance *
auth_instance_new(guint16 id)
{
    struct auth_instance *inst;
    struct auth_type *at;

    at = g_hash_table_lookup(auth_hash, &id);
    if (at == NULL)
        return NULL;

    inst = g_new0(struct auth_instance, 1);
    if (inst == NULL)
        return NULL;

    inst->at = at;
    at->refcount++;

    if (at->start_inst != NULL)
        (at->start_inst)(at, inst);

    return inst;
}

void
auth_instance_free(struct auth_instance *inst)
{
    struct auth_type *at = inst->at;

    if (at != NULL) {
        if (at->stop_inst != NULL)
            (at->stop_inst)(at, inst);

        at->refcount--;
#ifdef DEBUG
        if (at->refcount < 0)
            l_warn("auth_type (%p) refcount falls to %d!\n", at, at->refcount);
#endif
    }

    g_free(inst);
}

void
auth_register(struct auth_type *at)
{
#ifdef DEBUG
    if (g_hash_table_lookup(auth_hash, &at->auth_id) != NULL)
        l_warn("auth id %d already registered!\n", at->auth_id);
#endif

    l_debug("registering '%s' as auth type %d\n", at->name, at->auth_id);

    g_hash_table_insert(auth_hash, &at->auth_id, at);
}

void
auth_unregister(struct auth_type *at)
{
#ifdef DEBUG
    if (g_hash_table_lookup(auth_hash, &at->auth_id) == NULL)
        l_warn("auth id %d not registered!\n", at->auth_id);
#endif

    g_hash_table_remove(auth_hash, &at->auth_id);
}

static void
auth_shutdown_one(gpointer key, gpointer value, gpointer data)
{
    struct auth_type *at = (struct auth_type *)value;

    if (at->refcount != 0) {
        l_warn("unable to shutdown auth type %d: refcount == %d\n", at->auth_id,
               at->refcount);
    } else {
        (at->shutdown)(at);
    }
}

void
auth_shutdown(void)
{
    g_hash_table_foreach(auth_hash, auth_shutdown_one, NULL);
    g_hash_table_destroy(auth_hash);
}

int
auth_init(struct daemon *dae)
{
    if (auth_hash != NULL) {
        l_warn("auth_hash != NULL in auth_init\n");
        return 0;
    }

    auth_hash = g_hash_table_new(u16_hash, u16_equal);
    if (auth_hash == NULL) {
        l_warn("unable to allocate auth hash table\n");
        return 0;
    }

    return 1;
}
