/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <autoconf.h>

#include "cipher.h"

#ifndef USE_VULNERABILITY

#include "log.h"
#include "wire.h"
#include "util.h"
#include "keys.h"
#include "list.h"
#include "debug.h"

#define VALID_CIPHER_FUNCS(FOO) \
    ( (FOO) && (FOO)->desc && \
      (FOO)->encrypt && (FOO)->decrypt && (FOO)->schedule && \
      (FOO)->init && (FOO)->free )

struct cipher_funcs *ciphers[MAX_CIPHERS];

void
cipher_free(struct cipher *ciph)
{
    ciph->cf->free(ciph);
    free(ciph);
}

struct cipher * 
cipher_new(unsigned int type)
{
    struct cipher *ciph;
    struct cipher_funcs *cf;

    if ( type < 0 || type >= MAX_CIPHERS ) {
        l_debug("type %d is out of range of valid ciphers\n", type);
        return NULL;
    }

    cf = ciphers[type];

    /*
     * make sure the cipher is wholly registered.
     */
    if ( ! VALID_CIPHER_FUNCS(cf) ) {
        return NULL;
    }

    if ( (ciph = g_new0(struct cipher, 1)) == NULL) {
        return NULL;
    }

    ciph->type = type;
    ciph->cf = cf;

    if ( ciph->cf->init(ciph) ) {
        free(ciph);
        return NULL;
    }

    l_debug("allocated ciph %p of type %d\n", ciph, type);

    return ciph;
}

/*
 * XXX all of these little wrappers should export async apis so
 * that we can hand off to worky threads
 */
void
cipher_schedule(struct cipher *ciph, char *data, int len)
{
    ciph->cf->schedule(ciph, data, len);
}

void
cipher_encrypt(struct cipher *ciph, 
        void *in, int in_len, 
        void *out, int out_len)
{
    ciph->enc++;
    ciph->enc_bytes += in_len;

    if ( in_len )
        ciph->cf->encrypt(ciph, in, in_len, out, out_len);
}

void
cipher_decrypt(struct cipher *ciph, 
        void *in, int in_len, 
        void *out, int out_len)
{
    ciph->dec++;
    ciph->dec_bytes += in_len;

    if ( in_len )
        ciph->cf->decrypt(ciph, in, in_len, out, out_len);
}


/*
 * these next two functions duplicate too much code and would be way
 * cleaner if they could just return a new pkt.. 
 */
void
cipher_multi_encrypt( struct list_head *list,
        void *in, int in_len, 
        void *out, int out_len)
{
    struct list_head *pos, *n;
    struct packet *pkt;
    void *out_ptr, *in_ptr;

    if ( (pkt = pkt_getref_alloc(0, out_len)) == NULL )
        return; /* XXX :/ */

    pkt_push_tail(pkt, out_len);

    l_debug("%p,%d -> %p,%d -> %p,%d\n", 
            in, in_len, pkt->head, pkt->len, out, out_len);


    out_ptr = in;
    in_ptr = out;

    list_for_each_rev_safe(pos, n, list) {
        struct keys *keys = list_entry(pos, struct keys, multi_entry);

        SWAP(out_ptr, in_ptr);

        l_debug("%p => %p\n", in_ptr, out_ptr);

        cipher_encrypt(keys->out, 
                in_ptr, in_len,
                out_ptr, out_len);

        /*
         * after the initial input from the in buffer, start bouncing
         * between the output and tmp
         */
        if ( in_ptr == in )
            in_ptr = pkt->head;
    }

    /*
     * nuts, we did the wrong number so our output is sitting in the
     * tmp buffer
     */
    if ( out_ptr != out )
        memcpy(out, out_ptr, out_len);

    ref_release(pkt);
}

void
cipher_multi_decrypt( struct list_head *list,
        void *in, int in_len, 
        void *out, int out_len)
{
    struct list_head *pos, *n;
    struct packet *pkt;
    void *out_ptr, *in_ptr;

    if ( (pkt = pkt_getref_alloc(0, out_len)) == NULL )
        return; /* XXX :/ */

    pkt_push_tail(pkt, out_len);

    l_debug("%p,%d -> %p,%d -> %p,%d\n", 
            in, in_len, pkt->head, pkt->len, out, out_len);

    out_ptr = in;
    in_ptr = out;

    list_for_each_safe(pos, n, list) {
        struct keys *keys = list_entry(pos, struct keys, multi_entry);

        SWAP(out_ptr, in_ptr);

        l_debug("%p => %p\n", in_ptr, out_ptr);

        cipher_decrypt(keys->in, 
                in_ptr, in_len,
                out_ptr, out_len);

        if ( in_ptr == in )
            in_ptr = pkt->head;
    }

    if ( out_ptr != out )
        memcpy(out, out_ptr, out_len);

    ref_release(pkt);
}

static int
cipher_measure(unsigned int type)
{
    struct cipher *ciph;
#define BOG_KEY_LEN 64
    char *bogus_key;
#define BOG_LEN 1500
    char *in, *out;
#define BOG_REPS 2
    int i;
    double val;
    struct timeval start, stop;
    unsigned usecs, bytes;
    char extension[3];

    bogus_key = alloca(BOG_KEY_LEN);
    in = alloca(BOG_LEN);
    out = alloca(BOG_LEN);

    if ( !bogus_key || !in || !out )
        return -1;

    if ( (ciph = cipher_new(type)) == NULL )
        return -1;

    cipher_schedule(ciph, bogus_key, 128);

    usecs = 0;
    bytes = 0;

    for ( i = 0 ; i < BOG_REPS ; i++ ) {

        gettimeofday(&start, NULL);
        cipher_encrypt(ciph, in, BOG_LEN, out, BOG_LEN);
        gettimeofday(&stop, NULL);

        usecs += TV_USECS(&stop) - TV_USECS(&start);
        bytes += BOG_LEN;
    }

    pretty_rate(bytes, usecs, &val, extension);

    l_warn("%s encryption: %f %s/s\n", ciph->cf->desc, val, extension);

    usecs = 0;
    bytes = 0;

    for ( i = 0 ; i < BOG_REPS ; i++ ) {

        gettimeofday(&start, NULL);
        cipher_decrypt(ciph, in, BOG_LEN, out, BOG_LEN);
        gettimeofday(&stop, NULL);

        usecs += TV_USECS(&stop) - TV_USECS(&start);
        bytes += BOG_LEN;
    }

    pretty_rate(bytes, usecs, &val, extension);

    l_warn("%s decryption: %f %s/s\n", ciph->cf->desc, val, extension);

    cipher_free(ciph);

    return 0;
}

int 
cipher_register(unsigned int type, struct cipher_funcs *cf)
{
    struct cipher_funcs *our_cf;
    /*
     * types must be in wire.h before being registered.. thast a good way to
     * try and ensure that people realize that the types are exported to the
     * wire and need to be carefully allocated..
     */
    if ( type < 0 || type >= MAX_CIPHERS ) {
        l_debug("type %d is out of range of valid ciphers\n", type);
        return -1;
    }

    our_cf = ciphers[type];

    if ( VALID_CIPHER_FUNCS(our_cf) ) {
        /* XXX fastest cipher should win :) */
        l_debug("type %d is already registered\n", type);
        return -1;
    }

    ciphers[type] = cf;

    l_debug("registering '%s' as cipher type %d\n", 
            cf->desc, type); 

    cipher_measure(type);

    return 0;
}

void 
cipher_init(void) 
{
    memset(ciphers, 0, sizeof(ciphers));
}

#else

struct cipher omnicipher = {0, };

#endif
