/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *    (c) Copyright 2001 Phil Schwan <phil@off.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * config clients connect to our listening unix domain socket
 * and send us two ints.. a command, and a shared mem id.  at the
 * start of the shared mem region is a control struct that describes
 * the command.  if the server has a response to write, it can fill
 * in the struct.  the command return code is written back down the 
 * connected unix socket.  -errno for errors, 0 for ok, or the number
 * of bytes following the struct that is the return.. the semantics
 * of the response depend on the command, somewhat obviously.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#define UNIX_PATH_MAX 108 /* why doesn't glibc use a define in sys/un? */
#include <sys/un.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <glib.h>

#include "log.h"
#include "main.h"
#include "cmd.h"
#include "udp.h"
#include "link.h"
#include "tun.h"
#include "debug.h"
#include "cipher.h"
#include "config.h"

/*
 * we put all these monster files in here 'cause they'd just clutter
 * up the other files which have much more interesting code to
 * pay attention to..
 */

struct buf_ptr {
    void *ptr;
    int remaining;
};

static void
fill_cipher_stats(struct cipher_stats *cs, struct cipher *ciph)
{
    if ( ciph == NULL) {
        memset(cs, 0, sizeof(struct cipher_stats));
        return;
    }
    cs->enc = ciph->enc;
    cs->enc_bytes = ciph->enc_bytes;
    cs->dec = ciph->dec;
    cs->dec_bytes = ciph->dec_bytes;
}
static void
add_cipher_stats(struct cipher_stats *cs, struct cipher *ciph)
{
    if ( ciph == NULL)
        return;
    cs->enc += ciph->enc;
    cs->enc_bytes += ciph->enc_bytes;
    cs->dec += ciph->dec;
    cs->dec_bytes += ciph->dec_bytes;
}

static void 
link_show_one(gpointer key, gpointer data, gpointer user_data)
{
    struct buf_ptr *bp = user_data;
    struct link *link = data;
    struct cmd_link_show cls;

    cls.addr = link->sin.sin_addr.s_addr;
    cls.port = link->sin.sin_port;
    cls.id = ntohl(link->id);
    /*
     * XXX being a little fast and lose with the endianness definition
     * in the shm structs :/
     */
    cls.flags = ( link->up ? CMD_LINK_F_UP : 0) |
                ( link->neighbour ? CMD_LINK_F_NEI : 0) |
                ( link->neg_initiator ? CMD_LINK_F_INIT : 0);
    cls.tx_packets = link->seq.tx_seq - 1;
    cls.rx_packets = link->rx_packets;
    cls.missed_packets = link->seq.missed_packets;

    fill_cipher_stats(&cls.cs, link->left->out);
    add_cipher_stats(&cls.cs, link->right->out);
    add_cipher_stats(&cls.cs, link->left->in);
    add_cipher_stats(&cls.cs, link->right->in);
    if ( link->valid_new ) {
        add_cipher_stats(&cls.cs, link->left->out_new);
        add_cipher_stats(&cls.cs, link->right->out_new);
        add_cipher_stats(&cls.cs, link->left->in_new);
        add_cipher_stats(&cls.cs, link->right->in_new);
    }

    memcpy(bp->ptr, &cls, sizeof(cls));

    bp->ptr += sizeof(cls);
    bp->remaining -= sizeof(cls);
}

static int 
link_show(struct daemon *dae, void *buf, size_t len)
{
    struct buf_ptr bp = {
        ptr: buf, 
        remaining: len
    };
    GList *list;

    for ( list = dae->socket_list ; list ; list = g_list_next(list) ) {
        struct socket *sk = list->data;

        g_hash_table_foreach(sk->link_id_hash, link_show_one, &bp);
    }

    return len - bp.remaining;
}

static int lame_generation = 5;

static void 
tun_show_one(gpointer key, gpointer data, gpointer user_data)
{
    struct buf_ptr *bp = user_data;
    struct tunnel *tun = data;
    struct cmd_tunnel_show cts;

    /*
     * don't show dups
     */
    if ( tun->lame_show_generation == lame_generation )
        return;
    tun->lame_show_generation = lame_generation;

    cts.local_id_forward = htonl(tun->local_id_forward);
    cts.remote_id_forward = htonl(tun->remote_id_forward);
    cts.local_id_reverse = htonl(tun->local_id_reverse);
    cts.remote_id_reverse = htonl(tun->remote_id_reverse);
    cts.link_id_reverse = htonl(tun->link_reverse ? tun->link_reverse->id : 0);
    cts.link_id_forward = htonl(tun->link_forward ? tun->link_forward->id : 0);
    cts.addr = tun->addr;
    cts.netmask = tun->netmask;
    if ( tun->intf ) 
        strncpy(cts.dev, tun->intf->dev, IFNAMSIZ);
    else 
        strcpy(cts.dev, "NONE");

    cts.flags = ( tun->state.neg_initiator ? CMD_TUNNEL_F_INIT : 0) |
                ( tun->state.neg_forward ? CMD_TUNNEL_F_NEG_FWD : 0) |
                ( tun->state.saw_confack ? CMD_TUNNEL_F_NEG_FWD_ACK : 0);
    
    memcpy(bp->ptr, &cts, sizeof(cts));

    bp->ptr += sizeof(cts);
    bp->remaining -= sizeof(cts);
}

static void 
tun_show_link(gpointer key, gpointer data, gpointer user_data)
{
    struct buf_ptr *bp = user_data;
    struct link *link = data;

    g_hash_table_foreach(link->local_id_hash, tun_show_one, bp);
}

static int 
tun_show(struct daemon *dae, void *buf, size_t len)
{
    struct buf_ptr bp = {
        ptr: buf, 
        remaining: len
    };
    GList *list;

    lame_generation++;

    for ( list = dae->socket_list ; list ; list = g_list_next(list) ) {
        struct socket *sk = list->data;

        g_hash_table_foreach(sk->link_id_hash, tun_show_link, &bp);
    }

    return len - bp.remaining;
}

static int
tun_add(struct daemon *dae, void *buf, size_t len)
{
    struct cmd_tunnel_add *cta;
    struct link *link;
    int ret, cta_size;

    cta_size = sizeof(struct cmd_tunnel_add) + 3*sizeof(guint32);

    if ( len < cta_size )
        return -EINVAL;

    cta = alloca(cta_size);

    memcpy(cta, buf, cta_size);
    cta->id = ntohl(cta->id);

    l_debug("trying to find link %08x\n", cta->id);

    if ( cta->num_addrs < 0 )
        return -EINVAL;
    if ( cta->num_addrs > 2 )
        return -E2BIG;

    link = link_id_getref(dae->socket_list->data, cta->id);
    if ( link == NULL )
        return -ENOENT;

    ret = tunnel_create(link, cta->num_addrs, cta->addrs);

    ref_release(link);

    return ret;
}

static int
tun_del(struct daemon *dae, void *buf, size_t len)
{
    struct cmd_tunnel_del ctd;
    struct tunnel *tunnel = NULL;
    struct link *link;
    int ret;

    memcpy(&ctd, buf, sizeof(ctd));
    ctd.local_id = ntohl(ctd.local_id);
    ctd.link_id = ntohl(ctd.link_id);

    l_debug("tunnel link id %08x local id %08x\n", 
            ctd.link_id, ctd.local_id);

    link = link_id_getref(dae->socket_list->data, ctd.link_id);
    if ( link == NULL )
        return -ENOENT;

    ret = -ENOENT;
    tunnel = tunnel_id_getref(link, ctd.local_id);
    if ( tunnel == NULL ) 
        goto out;

    tunnel_teardown(tunnel, NULL);

    ret = 0;

    ref_release(tunnel);
out:
    ref_release(link);
    return ret;
}


/*
 * and the machanics for calling all the above functions..
 */
static int
cmd_process(struct daemon *dae, int cmd, int shm_id)
{
    void *shm_buf;
    struct shmid_ds ds;
    size_t size;
    int ret;

    if ( cmd < 0  || cmd >= CMD_MAX )
        RETURN(-EINVAL);

    if ( (shm_buf = shmat(shm_id, 0, 0)) == NULL )
        RETURN(-EINVAL);

    shmctl(shm_id, IPC_STAT, &ds);

    if ( ds.shm_segsz <= 0 ) {
        ret = -EINVAL;
        goto out;
    }

    size = ds.shm_segsz;

    switch(cmd) {
        case CMD_LINK_SHOW:
            ret = link_show(dae, shm_buf, size);
            break;
        case CMD_TUNNEL_SHOW:
            ret = tun_show(dae, shm_buf, size);
            break;
        case CMD_TUNNEL_ADD:
            ret = tun_add(dae, shm_buf, size);
            break;
        case CMD_TUNNEL_DEL:
            ret = tun_del(dae, shm_buf, size);
            break;
        case CMD_TOPO_SHOW:
            ret = topo_show(dae, shm_buf, size);
            break;
        default:
            ret = -ENOSYS;
            break;
    };

    l_debug("command %d returned %d\n", cmd, ret);

out:
    shmdt(shm_buf);
    RETURN(ret);
}

static gboolean 
pfd_prepare(GSource *source, gint *timeout)
{
    *timeout = -1; /* block forever */
    return FALSE;  /* we have to go to poll to find out if we're ready */
}

/* FIXME: GSource is technically marked private, but how else am I supposed to
 * do this?  They just mean ``don't _change_ stuff directly,'' right? --phil */
static gboolean 
pfd_check(GSource *source)
{
    GSList *l;

    for (l = source->poll_fds; l != NULL; l = l->next) {
        GPollFD *pfd = (GPollFD *)l->data;

        if (pfd->revents != 0)
            return 1;
    }

    return 0;
}

static gboolean 
pfd_dispatch(GSource *source, GSourceFunc callback, gpointer user_data)
{
    struct daemon *dae = (struct daemon *)user_data;
    GPollFD *pfd = &dae->cmd.pfd;
    struct sockaddr_un from;
    socklen_t fromlen = sizeof(struct sockaddr_un);
    int ret, len, args[2];

    len = RECVFROM(pfd->fd, args, sizeof(args), 0, &from, &fromlen);

    ret = -EINVAL;

    if ( len == sizeof(args) )
        ret = cmd_process(dae, args[0], args[1]);

    SENDTO(pfd->fd, &ret, sizeof(ret), 0, &from, fromlen);

    return TRUE;
}

static GSourceFuncs pfd_source_funcs = {
    prepare: pfd_prepare,
    check: pfd_check,
    dispatch: pfd_dispatch,
};

static void
cmd_shutdown(gpointer data)
{
    struct cmd_sock *cmd = data;

    g_source_remove_poll(cmd->source, &cmd->pfd);
    g_source_destroy(cmd->source);
    unlink(cmd->sun.sun_path);
}

int
cmd_init(struct daemon *dae)
{
    struct sockaddr_un *sun = &dae->cmd.sun;
    GPollFD *pfd = &dae->cmd.pfd;
    struct cmd_sock *cmd = &dae->cmd;
    struct stat st;

    if ( global_config->management_socket == NULL ) {
        l_critical("specify a management socket in the config\n");
        return 0;
    }

    sun->sun_family = AF_LOCAL;
    snprintf(sun->sun_path, UNIX_PATH_MAX-1, "%s.%d", 
            global_config->management_socket, getpid());
    sun->sun_path[UNIX_PATH_MAX-1] = '\0';

    if ( stat(sun->sun_path, &st) == 0) 
        unlink(sun->sun_path);

    if ( (pfd->fd = socket(AF_LOCAL, SOCK_DGRAM, 0)) < 0 ) {
        l_critical("coulnd't initialize command unix socket\n");
        return 0;
    }

    if ( BIND(pfd->fd, sun, sizeof(struct sockaddr_un))) {
        l_critical("coulnd't bind to unix socket path '%s'\n", sun->sun_path);
        goto err_close;
    }

    l_debug("bound to unix %s\n", sun->sun_path);

    pfd->events = G_IO_IN | G_IO_HUP | G_IO_ERR;

    cmd->source = g_source_new(&pfd_source_funcs, sizeof(GSource));
    if (cmd->source == NULL) {
        l_critical("unable to allocate GSource\n");
        goto err_close;
    }

    if ( hook_register(&dae->shutdown_hooks, cmd_shutdown, cmd) )
        goto err_source;

    g_source_add_poll(cmd->source, pfd);
    g_source_set_priority(cmd->source, G_PRIORITY_DEFAULT);
    g_source_set_callback(cmd->source, NULL, dae, NULL);
    g_source_attach(cmd->source, NULL);

    return 1;

err_source:
    g_source_destroy(cmd->source);

err_close:
    close(pfd->fd);

    return 0;
}
