/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *    (c) Copyright 2001 Phil Schwan <phil@off.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * yet another parser.  its our calling!
 */

#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <glib.h>

#include "config.h"
#include "main.h"
#include "log.h"
#include "defaults.h"
#include "udp.h"
#include "link.h"

enum {
    CT_SIN,
    CT_INT,
    CT_ULONG,
    CT_SINC_LIST,
    CT_STRING,
};

#define CF_OFF(foo) G_STRUCT_OFFSET(struct config, foo)

struct config_element {
    char *name;
    char *attr;
    int type;
    int offset;
};

struct config_tree_node {
    char *name;
    struct config_element *elements;
    struct config_tree_node *parent;
    struct config_tree_node *children[0];
};

struct config_tree_node root_node;

struct config_element globals_elements[] = {
    { "plugin", "directory", CT_STRING, CF_OFF(plugin_directory)},
    { "management_socket", "path", CT_STRING, CF_OFF(management_socket)},
    { "debug", "level", CT_INT, CF_OFF(debug)},
    { NULL, }
};
struct config_tree_node globals_node = {
    name:       "globals", 
    elements:   globals_elements,
    parent:     &root_node,
    children:   {NULL },
};

struct config_element cebolla_elements[] = {
    { "udp_listen", "ipv4", CT_SIN, CF_OFF(udp_listen)},
    { "link", "ipv4", CT_SINC_LIST, CF_OFF(link_sinc_head)},
    { "lsa_file", "path", CT_STRING, CF_OFF(lsa_file)},
    { "log_file", "path", CT_STRING, CF_OFF(log_file)},
    { "lsa_retrans_timeout", "seconds", CT_ULONG, CF_OFF(lsa_retrans_timeout)},
    { "lsa_expiry_timeout", "seconds", CT_ULONG, CF_OFF(lsa_expiry_timeout)},
    { "lsa_refresh_interval", "seconds", CT_ULONG, CF_OFF(lsa_refresh_interval)},
    { "neg_recalc_timeout", "seconds", CT_ULONG, CF_OFF(neg_recalc_timeout)},
    { "ip_pool", "network", CT_SINC_LIST, CF_OFF(ip_pool_sinc_head)},
    {"pgp_certificate", "path", CT_STRING, CF_OFF(pgp_certificate)},
    { NULL, }
};
struct config_tree_node cebolla_node = {
    name:       "cebolla", 
    elements:   cebolla_elements,
    parent:     &root_node,
    children:   {NULL },
};

struct config_element root_elements[] = {
    { NULL, },
};
struct config_tree_node root_node = {
    name:       "root", 
    elements:   root_elements,
    parent:     NULL,
    children:   {&globals_node, &cebolla_node, NULL},
};

struct parse_state {
    struct daemon *dae;
    struct config *config;
    struct config_tree_node *cur_node;
};

struct sin_container {
    struct list_head entry;
    struct sockaddr_in sin;
};

struct config *global_config = NULL;

struct config default_config = {
    lsa_retrans_timeout:    1,
    lsa_expiry_timeout:     3,
    lsa_refresh_interval:   60,
    neg_recalc_timeout:     30,
};

static char *ip_port_delims = ":/";

static guint32
parse_ip(const char *str)
{
    guint32 ret = INADDR_ANY;
    char *ip = strdup(str);
    char *sep = strpbrk(ip, ip_port_delims);

    if ( sep )
        *sep = '\0';

    if ( sep == ip )
        return INADDR_ANY;

    ret = inet_addr(ip);
    if (ret == INADDR_NONE) {
        l_error("Invalid IP address '%s'\n", ip);
    }

    free(ip);
    return ret;
}

static guint16
parse_port(const char *str)
{
    guint32 ret = CEBOLLA_PORT;
    char *sep = strpbrk(str, ip_port_delims);

    if ( sep == NULL )
        goto out;

    sep++;

    if ( *sep == '\0' )
        goto out;

    if ( ! isdigit(*sep) )
        l_error("bad port %s\n", sep);

    ret = strtoul(sep, NULL, 10);

out:
    return htons(ret);
}

static void
parser_start(GMarkupParseContext *context, const gchar *elem,
             const gchar **attr_name, const gchar **attr_val,
             gpointer user_data, GError **error)
{
    struct parse_state *ps = user_data;
    struct config_tree_node *ctn;
    struct config_element *cele;
    void *ptr;
    int i;

    for ( i = 0; ( ctn = ps->cur_node->children[i] ) ; i++) {
        if ( !strcmp(elem, ctn->name) ) {
            l_debug("descending into <%s and making a new config\n", elem);
            ps->cur_node = ctn;

            ps->config = g_new0(struct config, 1);
            memcpy(ps->config, &default_config, sizeof(struct config));

            INIT_LIST_HEAD(&ps->config->link_sinc_head);
            INIT_LIST_HEAD(&ps->config->ip_pool_sinc_head);
            return;
        }
    }

    for ( cele = ps->cur_node->elements ; cele->name ; cele++ ) {
        if ( ! strcmp(cele->name, elem) )
                break;
    }

    if ( cele->name == NULL ) {
        l_warn("unknown element: <%s>, ignoring..\n", elem);
        return;
    }

    for ( i = 0; attr_name[i] != NULL ; i++ ) {
        if ( ! strcmp(cele->attr, attr_name[i]) )
            break;
    }

    if ( cele->attr && attr_name[i] == NULL ) {
        l_error("'%s=' attribute required for <%s>\n", cele->attr, elem);
    }

    l_debug("parsing <%s %s=%s\n", elem, attr_name[i], attr_val[i]);

    ptr = G_STRUCT_MEMBER_P(ps->config, cele->offset);

    switch(cele->type) {

        case CT_STRING: 
            {
                char **str = ptr;
                *str = g_strdup(attr_val[i]);
            }
            break;

        case CT_ULONG:
            {
                unsigned long *val = ptr;

                *val = strtoul(attr_val[i], NULL, 10);
            }
            break;

        case CT_INT:
            {
                int *val = ptr;

                *val = atoi(attr_val[i]);
            }
            break;

        case CT_SIN:
            {
                struct sockaddr_in *sin = ptr;

                sin_init(sin);
                sin->sin_port = parse_port(attr_val[i]);
                sin->sin_addr.s_addr = parse_ip(attr_val[i]);
            }
            break;

        case CT_SINC_LIST:
            {
                struct list_head *list = ptr;
                struct sin_container *sinc = g_new0(struct sin_container, 1);

                if ( sinc == NULL ) 
                    l_error("out of memoery parsing ip list\n");

                sin_init(&sinc->sin);
                sinc->sin.sin_port = parse_port(attr_val[i]);
                sinc->sin.sin_addr.s_addr = parse_ip(attr_val[i]);
                list_add(&sinc->entry, list);
            }
            break;
    }
}

static void
parser_end(GMarkupParseContext *context, const gchar *elem,
            gpointer user_data, GError **error)
{
    struct parse_state *ps = user_data;

    if ( ! strcmp(elem, ps->cur_node->name) ) {
        l_debug("descending back from %s to %s\n",  
                ps->cur_node->name, ps->cur_node->parent->name);
        ps->cur_node = ps->cur_node->parent;
    }

    /*
     * some special casing :/
     */
    if ( ! strcmp(elem, "globals") ) {
        if ( global_config ) 
            l_error("there can be only one <globals> section\n");
        global_config = ps->config;
        ps->config = NULL;
    }

    if ( ! strcmp(elem, "cebolla") ) {
        struct socket *sk;
        struct list_head *pos, *n;

        sk = socket_prepare(ps->dae, &ps->config->udp_listen);
        if ( sk == NULL )
            l_error("ran out of memory setting up a socket\n");

        list_for_each_safe(pos, n, &ps->config->ip_pool_sinc_head) {
            struct sin_container *sinc = list_entry(pos,
                                            struct sin_container, entry);
            /* XXX all this ip block shit needs to be somewhere else */
            struct ip_block *ipb = g_new0(struct ip_block, 1);

            if ( ipb == NULL )
                l_error("couldn't allocate ip block\n");
            if ( sinc->sin.sin_addr.s_addr == INADDR_ANY ) 
                l_error("<ip_pool> neeeds a network\n");
            sinc->sin.sin_port = ntohs(sinc->sin.sin_port);
            if ( sinc->sin.sin_port > 32 )
                l_error("<ip_pool> network width can't be > 32\n");

            ipb->base = ntohl(sinc->sin.sin_addr.s_addr);
            ipb->num_addrs = 1 < ( 32 - sinc->sin.sin_port );
            ipb->num_words = ipb->num_addrs / (sizeof(ipb->mask[0]) * 8) + 1;
            ipb->mask = malloc(ipb->num_words * sizeof(ipb->mask[0]));
            memset(ipb->mask, 0xff, ipb->num_words * sizeof(ipb->mask[0]));
            sk->ipb = ipb;

            list_del(&sinc->entry);
            free(sinc);
            break;
        }

        list_for_each_safe(pos, n, &ps->config->link_sinc_head) {
            struct sin_container *sinc = list_entry(pos,
                                            struct sin_container, entry);
            link_prepare(sk, &sinc->sin);

            list_del(&sinc->entry);
            free(sinc);
        }

        /* XXX this is what would create multiple daemons */
        ps->dae->config = ps->config;
        ps->config = NULL;
    }
}

int 
config_parse(struct daemon *dae, char *filename)
{
    GMarkupParseContext *ctx;
    GMarkupParser parser = {
        start_element: parser_start,
        end_element: parser_end
    };
    int fd;
    GError *err = NULL;
    void *ptr;
    struct stat st;
    struct parse_state ps;

    fd = open(filename, O_RDONLY);
    if (fd < 0)  {
        l_critical("Cannot open '%s': %s\n", filename, strerror(errno));
        return 0;
    }

    if ( fstat(fd, &st) < 0 ) {
        close(fd);
        l_critical("Cannot stat '%s': %s\n", filename, strerror(errno));
        return 0;
    }

    ptr = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0);

    close(fd);

    if ( ptr == MAP_FAILED ) {
        l_critical("Cannot mmap() '%s': %s\n", filename, strerror(errno));
        return 0;
    }

    ps.dae = dae;
    ps.cur_node = &root_node;

    ctx = g_markup_parse_context_new(&parser, 0, &ps, NULL);

    g_markup_parse_context_parse(ctx, ptr, st.st_size, &err);
    g_markup_parse_context_free(ctx);

    munmap(ptr, st.st_size);

    return 1;
}

