/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "log.h"
#include "digest.h"
#include "wire.h"

#define VALID_DIGEST_FUNCS(FOO) \
    ( (FOO) && (FOO)->desc && \
      (FOO)->update && (FOO)->complete && \
      (FOO)->init && (FOO)->free )

struct digest_funcs *digests[MAX_DIGESTS];

void
digest_free(struct digest *dig)
{
    dig->df->free(dig);
    free(dig);
}

struct digest * 
digest_new(unsigned int type)
{
    struct digest *dig;
    struct digest_funcs *df;

    if ( type < 0 || type >= MAX_DIGESTS ) {
        l_debug("type %d is out of range of valid digests\n", type);
        return NULL;
    }

    df = digests[type];

    /*
     * make sure the digest is wholly registered.
     */
    if ( ! VALID_DIGEST_FUNCS(df) ) {
        return NULL;
    }

    if ( (dig = g_new0(struct digest, 1)) == NULL) {
        return NULL;
    }

    dig->type = type;
    dig->df = df;

    if ( dig->df->init(dig) ) {
        free(dig);
        return NULL;
    }

    l_debug("allocated dig %p of type %d\n", dig, type);

    return dig;
}

/*
 * XXX all of these little wrappers should export async apis so
 * that we can hand off to worky threads
 */
void
digest_update(struct digest *dig, void *data, int len)
{
    dig->df->update(dig, data, len);
}

/*
 * what a bizarre function.  we want to return as much hash material
 * as the caller wants, but their hash function may not output something
 * that large.  so we'll hash the output of the hash until we reach
 * the amount of bytes they want.
 */
void
digest_complete(struct digest *dig, void *caller, int out_len)
{
    char *local;
    int len = 0; /* sigh */
    char *out = caller;

    local = alloca(dig->max_output);

    while (out_len > 0) {

        if ( out != caller ) {
            /* if we're filling in caller with extra ->complete()s we
             * have to seed the hash with the output of previous
             * runs.
             */
            dig->df->update(dig, out - len, len);
        }

        if ( out_len < dig->max_output) {
            /* 
             * if the caller didn't make room for the digest, we output
             * locally and copy..
             */
            len = dig->df->complete(dig, local);
            memcpy(out, local, MIN(len, out_len));
        } else  {
            /* great, we can write directly to them.. */
            len = dig->df->complete(dig, out);
        }

        out_len -= len;
        out += len;
    }

    return;
}

int 
digest_register(unsigned int type, struct digest_funcs *df)
{
    struct digest_funcs *our_df;
    /*
     * types must be in wire.h before being registered.. thast a good way to
     * try and ensure that people realize that the types are exported to the
     * wire and need to be carefully allocated..
     */
    if ( type < 0 || type >= MAX_DIGESTS ) {
        l_debug("type %d is out of range of valid digests\n", type);
        return -1;
    }

    our_df = digests[type];

    if ( VALID_DIGEST_FUNCS(our_df) ) {
        /* XXX fastest digest should win :) */
        l_debug("type %d is already registered\n", type);
        return -1;
    }

    digests[type] = df;

    l_debug("registering '%s' as digest type %d\n", df->desc, type); 

    return 0;
}

void 
digest_init(void) 
{
    memset(digests, 0, sizeof(digests));
}
