/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * the code path for incoming packets..
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "log.h"
#include "main.h"
#include "pkt.h"
#include "udp.h"
#include "topo.h"
#include "link.h"
#include "debug.h"
#include "cipher.h"
#include "tunnel.h"
#include "outgoing.h"

/*
 * discover if either of our key generations can decrypt this 
 * packet..
 */
static int 
check_link_headers( struct packet *in_pkt, struct w_link_hdr *lh, 
        struct cipher *left, struct cipher *right)
{
    struct w_link_hdr left_lh;

    cipher_decrypt(right,
            in_pkt->head, sizeof(struct w_link_hdr),
            lh, sizeof(struct w_link_hdr));

    pkt_push_head(in_pkt, sizeof(struct w_link_hdr));

    cipher_decrypt(left,
            in_pkt->head, sizeof(struct w_link_hdr),
            &left_lh, sizeof(struct w_link_hdr));

    pkt_pull_head(in_pkt, sizeof(struct w_link_hdr));

    if ( memcmp(&left_lh, lh, sizeof(struct w_link_hdr)) == 0) 
        return 1;

    return 0;
}

static int 
try_link_hdr_decrypt(struct link *link, struct packet *in_pkt, 
        struct w_link_hdr *lh)
{
    if ( check_link_headers(in_pkt, lh, link->left->in, link->right->in) ) {

        link_rekey_action( link,
                keys_rekey_dec_notice(&link->rek,  REK_PRIMARY) );

        return 1;
    }

    if ( link->valid_new &&
        check_link_headers(in_pkt, lh, link->left->in_new, link->right->in_new)) {

        link_rekey_action( link,
                keys_rekey_dec_notice(&link->rek,  REK_SECONDARY) );
        return 1;
    }

    return 0;
}

/*
 * a packet is coming in to the device for this tunnel..
 */
void
incoming_frame(struct tunnel *tunnel, struct packet *pkt)
{
    guint32 id = 0;
    struct link *link;
    /*
     * XXX could apply bfd here...
     */
    hexdump(pkt->head, pkt->len, "incoming ip?\n");

    if ( tunnel->remote_id_forward ) {
        id = tunnel->remote_id_forward;
        link = tunnel->link_forward;
    } else {
        id = tunnel->remote_id_reverse;
        link = tunnel->link_reverse;
    }

    if ( id == 0 ) {
        l_debug("incoming ip frame to a tunnel with no remote id?\n");
        return;
    }

    outgoing_tunnel_msgend(tunnel, link, pkt, MSG_TYPE_IPV4, id);
}

/*
 * process it only if we're the endpoint.  if not, send it on
 */
static void
incoming_tunnel_msgend(struct tunnel *tunnel, struct packet *in_pkt, guint32 id)
{
    struct w_tun_hdr th;
    struct link *link;
    struct packet *pkt;

    if ( in_pkt->len < sizeof(th) )
        return;

    if ( (pkt = pkt_getref_alloc(0, in_pkt->len)) == NULL )
        return;

    pkt_push_tail(pkt, in_pkt->len);

    if ( id == tunnel->local_id_forward ) {
        /* coming for our forward id, so its going in reverse direction */
        id = tunnel->remote_id_reverse;
        link = tunnel->link_reverse;
    } else {
        /* coming for our reverse id, so going forward.. */
        id = tunnel->remote_id_forward;
        link = tunnel->link_forward;
    }

    if ( !tunnel->state.neg_initiator && id == tunnel->remote_id_reverse ) 
        cipher_multi_encrypt( &tunnel->multi_keys_head, 
                                in_pkt->head, in_pkt->len,
                                pkt->head, pkt->len);
    else 
        cipher_multi_decrypt( &tunnel->multi_keys_head, 
                                in_pkt->head, in_pkt->len,
                                pkt->head, pkt->len);

    memcpy(&th, pkt->head, sizeof(th));
    th.seq = ntohs(th.seq);
    pkt_push_head(pkt, sizeof(th));

    /*
     * if neg_forward is set then we're a node in the middle of 
     * a path relaying negotiation packets.  We start doing this
     * when we have the reverse side negotiated, and until we
     * see a real packet come from the initiator after the responder
     * sends a confack.
     */
    if ( tunnel->state.neg_forward ) {

        if ( th.type == MSG_TYPE_NEG ) {
            tunnel_forward_neg(tunnel, pkt);
            goto out;

        } else  {

            if ( ! tunnel->state.saw_confack ) {
                l_debug("someone sent type %d to a forwarding tunnel\n", 
                        th.type);
                goto out;
            }

            tunnel->state.neg_forward = 0;
            tunnel->state.saw_confack = 0;

            g_hash_table_remove(tunnel->link_forward->neg_fwd_hash, 
                    tunnel->forwarding_init_id);

            tunnel->state.have_fwd_init_id = 0;
        } 
    }

    /*
     * if we have an id, we should be forwarding it along..
     */
    if ( id && ! tunnel->state.neg_forward) {
        l_debug("forwarding to %x on link %08x\n", id, link->id);
        pkt_pull_head(pkt, sizeof(th));
        outgoing_link(pkt, link, LINK_TUNNEL_MSGEND, id);
        goto out;
    }

    if ( th.type >= __MSG_TYPE_LAST ) {
        l_debug("someone sent weird type %d\n", th.type);
        goto out;
    }

    if ( seq_process(&tunnel->end_seq, th.seq) ) 
        goto out;

    switch ( th.type ) {

        case MSG_TYPE_IPV4: 
            if ( tunnel->intf == NULL ) {
                l_debug("got to an endpoint without an interface?\n");
                goto out;
            }

            int_deliver(tunnel->intf, pkt);
            break;

        case MSG_TYPE_NEG:
            tunnel_process_neg(tunnel->link_forward, pkt, 
                    tunnel->remote_id_forward);
            break;

        default:
            l_debug("shouldn't get here with type %d\n", th.type);
            break;
    }

out:
    ref_release(pkt);
}

/*
 * process an incoming packet.  its quite legal for people deeper in
 * the call-chain to grab a reference to the packet, so we don't mess
 * with it after calling them.  the link refcounting is to protect
 * everyone deeper in the stack from having it pulled out from 
 * under them..
 */
void
incoming_link(struct socket *sk, struct packet *in_pkt, 
        struct sockaddr_in *from)
{
    struct link *link;
    struct tunnel *tunnel = NULL;
    struct packet *pkt = NULL;
    struct w_link_hdr _lh, *lh = &_lh;

    hexdump(in_pkt->head, in_pkt->len, "------------- %s:%d\n",
            inet_ntoa(from->sin_addr), ntohs(from->sin_port));

    /*
     * if we don't know of this link, it might be link negotiation..
     */
    if ( (link = link_nei_getref(sk, from)) == NULL) {
        link_process_neg(sk, in_pkt, from);
        return;
    }

    /*
     * "You must be this tall... "
     */
    if (in_pkt->len < (2 * sizeof(struct w_link_hdr)) )
        goto out;

    if ( ! try_link_hdr_decrypt(link, in_pkt, lh) ) {
        /*
         * XXX might be a retransmitted confreq
         */
        l_debug("couldn't find valid link headers\n");
        goto out;
    }

    pkt_push_head(in_pkt, 2 * sizeof(struct w_link_hdr));

    /* lets just get these out of the way */
    lh->seq = ntohs(lh->seq);
    lh->id = ntohl(lh->id);

    if ( seq_process(&link->seq, lh->seq) )
        goto out;

    if ( lh->seq >= SEQ_REKEY_WATERMARK ) 
            link_rekey_action( link, keys_rekey_initiate(&link->rek) );

    link->rx_packets++;

    link->activity = time(NULL);

#if 0 /* bring me back when we enc/mac */
    /*
     * deal with the encrypted payload
     */
    if ( lh->encrypted > in_pkt->len ) {
        l_debug("link header enc %d is greater than pkt->len %d\n",
                lh->encrypted, in_pkt->len);
        goto out;
    }
    if ( (pkt = pkt_getref_alloc(0, in_pkt->len)) == NULL) {
        l_debug("couldn't allocate pkt header\n");
        goto out;
    }

    pkt_push_tail(pkt, lh->encrypted);

    cipher_decrypt(link->left.in,
            in_pkt->head, lh->encrypted, 
            pkt->head, lh->encrypted);

    pkt_push_head(in_pkt, lh->encrypted);
    /*
     * make room in our packet for the remaining clear bits of the
     * incoming packet and copy them over
     */
    memcpy( pkt_push_tail(pkt, in_pkt->len),  in_pkt->head,  in_pkt->len);
#endif

    pkt = in_pkt;

    /*
     * this may seem strange, but its to stop us from flooding topology
     * packets to a neighbour when we're the responder in a link
     * negotiation.  we don't want to flood them with topo stuff
     * until we really know that they got the confack
     */
    if ( !link->up )
        link_up(link);

    if ( lh->id )
        tunnel = tunnel_id_getref(link, lh->id);

    l_debug("incoming link type %d on link %08x (tunnel %p)\n", 
            lh->type, link->id, tunnel);

    switch (lh->type) {

        case LINK_REKEY_REQUEST:
            link_rekey_action( link,
                    keys_rekey_request(&link->rek, link->neg_initiator) );
            break;

        case LINK_ECHO_REQUEST:
            l_debug("got echo request\n");
            link_send_message(link, LINK_ECHO_REPLY);
            break;

        case LINK_ECHO_REPLY:
            l_debug("got echo reply\n");
            break;

        case LINK_LSA:
        case LINK_LSA_ACK:
            topo_incoming(pkt, link, lh->type);
            break;

        case LINK_TUNNEL_NEG:
            tunnel_process_neg(link, pkt, lh->id);
            break;

        case LINK_TUNNEL_MSGEND:
            if ( tunnel ) 
                incoming_tunnel_msgend(tunnel, pkt, lh->id);
            break;

        case LINK_TUNNEL_DESTROY:
            if ( tunnel ) 
                tunnel_teardown(tunnel, link);
            break;

        case LINK_DESTROY:
                link_teardown(link, link);
            break;

        default:
            l_debug("oh, my, unknown link type %d (0x%02x\n", 
                    lh->type, lh->type);
            break;
    };

out:
#if 0
    if ( pkt ) 
        pkt_decref(pkt);
#endif

    if ( tunnel )
        ref_release(tunnel);

    ref_release(link);
    return;
}
