/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "log.h"
#include "keys.h"
#include "util.h"
#include "cipher.h"
#include "wire.h"

static struct digest *keys_dig;

void
keys_destroy(struct keys *keys)
{
    if (keys->in)  
        cipher_free(keys->in);
    if (keys->out)  
        cipher_free(keys->out);
    if (keys->in_new)  
        cipher_free(keys->in_new);
    if (keys->out_new)  
        cipher_free(keys->out_new);

    free(keys);
}

/*
 * derives the next set of keys from the current keys and the
 * shared secret.  This is not reversable.  Each call to this will
 * produce a different set of keys.. different entities trying to keep
 * their key derivations in sync must be very careful to only call
 * this in coordination with whoever they're trying to be in sync
 * with.
 */
void
keys_derive(struct keys *keys)
{
    if ( ! (keys && keys->session) ) {
        l_warn("some joker passed us an uninitialized set of keys.\n");
        return;
    }

    digest_update(keys_dig, keys->md, MAX_DIGEST_BYTES);
    digest_update(keys_dig, keys->session->data, keys->session->len);
    digest_complete(keys_dig, keys->md, MAX_DIGEST_BYTES);

    cipher_schedule(keys->in_new, keys->md, MAX_DIGEST_BYTES);

    digest_update(keys_dig, keys->md, MAX_DIGEST_BYTES);
    digest_update(keys_dig, keys->session->data, keys->session->len);
    digest_complete(keys_dig, keys->md, MAX_DIGEST_BYTES);

    cipher_schedule(keys->out_new, keys->md, MAX_DIGEST_BYTES);

    if ( keys->ordering ) 
        SWAP(keys->in_new, keys->out_new);
}

/*
 * This is called to initialize 'line' of keys.  the string just adds some
 * noise to the shared secret.. it makes it so lots of keys intitially derived
 * from the same shared secret will be different.  both sides of a conversation
 * should use the same string with initially deriving keys..
 *
 * 'ordering' is a boolean that has to differ between either side of our
 * communications that use differnet keys for incoming and outgoing messages.
 * one sides 'in' is the others 'out'..
 */
struct keys *
keys_setup(struct secret *sec, int cipher_type,
        int ordering, char *string)
{
    struct keys *keys;

    if ( sec == NULL ) {
        l_warn("some naughty code path gave us a null sec\n");
        return NULL;
    }

    if ( (keys = g_new0(struct keys, 1)) == NULL )
        goto error;

    keys->ordering = ordering ? 1 : 0;
    keys->session = sec;

    keys->in = cipher_new(cipher_type); 
    keys->out = cipher_new(cipher_type); 
    keys->in_new = cipher_new(cipher_type); 
    keys->out_new = cipher_new(cipher_type); 

    if ( !keys->in || !keys->out || !keys->in_new || !keys->out_new) 
        goto error;

    /*
     * put some caller specified noise in the md 
     * to add some randomness to the initial derivation..
     */
    digest_update(keys_dig, string, strlen(string));
    digest_complete(keys_dig, keys->md, MAX_DIGEST_BYTES);

    keys_derive(keys);

    /*
     * we want the new keys immediately.
     */
    SWAP(keys->in, keys->in_new);
    SWAP(keys->out, keys->out_new);

    return keys;

error:
    if ( keys )
        keys_destroy(keys);

    return NULL;
}

/*
 * rekeying is a touchy state machine, we we keep all the code for doing it
 * here..
 *
 * -------
 *
 * calculates secondary keys
 * sends rekey request with old keys
 *
 *                              calculates secondary keys
 *                              fkips keys
 *                              sends dummy (ping) with new keys
 *
 * notices decryption with secondary
 * fkips keys, destroys secondary
 * marks rekeying as done
 * sends dummy (ping) with new keys
 *
 *                              notices decryption with primary
 *                              destroys secondary
 *                              marks rekeying as done
 * -------
 *
 * we destroy the old keysets once we see decryption suceed with the
 * new keysets.  this means that frames that are reordered around
 * the switch to the new keys may be lost (especially under linux 2.4 
 * softnet with its reordering of frames on smp?)
 */

/*
 * see keys.h for some clarification of these functions..
 */
int keys_rekey_initiate(struct rekeying *rek)
{
    if ( rek->in_progress )
        return 0;

    rek->in_progress = 1;
    rek->initiator = 1;

    return REK_DERIVE|REK_REQUEST;
}
int keys_rekey_dec_notice(struct rekeying *rek, int generation)
{
    if ( ! rek->in_progress )
        return 0;

    if ( rek->initiator && (generation == REK_SECONDARY)) {
        rek->in_progress = 0;
        return REK_FLIP|REK_DESTROY|REK_SEQRESET|REK_SEND;
    }

    if ( !rek->initiator && (generation == REK_PRIMARY)) {
        rek->in_progress = 0;
        return REK_DESTROY|REK_SEQRESET;
    }

    return 0;
}
int keys_rekey_request(struct rekeying *rek, int initiator)
{
    if ( ! rek->in_progress ) {
        rek->in_progress = 1;
        rek->initiator = 0;
        return REK_DERIVE|REK_FLIP|REK_SEND;
    }

    if ( rek->initiator ) {
        /*
         * both ends of the rekeying initiated at the same time. we
         * decide who remains the initiator based on the roles each
         * played in the negotiation of the higher level resource..
         */
        if ( ! initiator ) {
            l_debug("falling over to being a rekeying responder\n");
            rek->initiator = 0;
            /*
             * we've already derived our secondary keys, but now that
             * we're not the initiator we want to flip the keys..
             */
            return REK_FLIP|REK_SEND;
        }
    }

    return 0;
}

int
keys_init(void)
{
    if ( (keys_dig = digest_new(DIGEST_MD5)) == NULL) {
        l_warn("no keys_dig?!  Very sad.\n");
        return 0;
    }

    return 1;
}
