#ifndef _KEYS_H_
#define _KEYS_H_

#include "secret.h"
#include "digest.h"
#include "list.h"

struct keys {
	struct list_head multi_entry;
        /*
         * the session secret that these keys are derived from.
         * currently session secrets don't change for the lifetime
         * of a session, but that may change.  for now one must
         * destroy all the keys before destroying the secret they
         * came from.
         */
        struct secret *session;

        int ordering;

        char md[MAX_DIGEST_BYTES];

        /*
         * primary keys that should be used for input and output
         */
        struct cipher *in;
        struct cipher *out;
        /*
         * keys are derived into these, callers swap between these
         * and the full in/out as their protocols require.  we also
         * use these _new keys specifically when we're trying to find
         * out if the other side has derived their new set yet..
         */
        struct cipher *in_new;
        struct cipher *out_new;
};

int keys_init(void);

struct keys * keys_setup(struct secret *sec, int cipher_type,
                int ordering, char *string);
void keys_destroy(struct keys *keys);

void keys_derive(struct keys *keys);

/*
 * ------------------------------------------------------------
 * utility code for use by a layer who wants to rekey..
 * ------------------------------------------------------------
 */
struct rekeying {
        int in_progress;
        int initiator;
};

/*
 * the helper functions return bitmasks of actions to take... they
 * must be acted on in the order that they're defined here.
 */
enum {
        /*
         * derive your secondary set of keys with keys_derive();
         */
        REK_DERIVE      =       0x0001,
        /*
         * send a REKEY_REQUEST message over your medium
         */
        REK_REQUEST     =       0x0002,
        /*
         * swap the primary and secondary keysets..
         */
        REK_FLIP        =       0x0004,
        /*
         * destroy the secondary keyset.  you no longer need to try to
         * decrypt with it or report dec notices..
         */
        REK_DESTROY     =       0x0008,
	/*
	 * set the sequence number back to its initial value.
	 */
	REK_SEQRESET	=	0x0010,
        /*
         * send any message over the medium with the primary key set..
         */
        REK_SEND        =       0x0020
};

/*
 * called to initiate rekeying 
 */
int keys_rekey_initiate(struct rekeying *rek);

/*
 * called to tell the rekeying core which keys decrypted an incoming
 * packet.  this must be called and its actions taken before the
 * actual rekeying request message that may be in the packet is parsed.
 */
enum {
        /* generations.. */
        REK_PRIMARY,
        REK_SECONDARY
};
int keys_rekey_dec_notice(struct rekeying *rek, int generation);


/*
 * called on receipt of a rekeying request packet.  'initiator' specifies
 * whether or not the caller was the initiator of the higher level
 * medium that the request came in over.. 
 */
int keys_rekey_request(struct rekeying *rek, int initiator);

#endif
