/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * link management.  negotiating links, making sure they're still alive,
 * tearing them down, etc.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <glib.h>

#include "main.h"
#include "util.h"
#include "udp.h"
#include "topo.h"
#include "link.h"
#include "pkt.h"
#include "wire.h"
#include "log.h"
#include "neg.h"
#include "seqwind.h"
#include "outgoing.h"
#include "list.h"
#include "tunnel.h"

static void
link_destroy(struct link *link)
{
    l_debug("destroying link %08x at %p\n", link->id, link);

    if ( link->left )
        keys_destroy(link->left);
    if ( link->right )
        keys_destroy(link->right);
    if ( link->session )
        secret_free(link->session);
    g_hash_table_destroy(link->neg_fwd_hash);
    g_hash_table_destroy(link->local_id_hash);
    free(link);
}

/*
 * effectively remove a link from operation.  the hash removals
 * will stop others from getting new references to it.  the loss
 * of th references in the hashes could also destroy the link if
 * no one else isusing it..
 */
void
link_teardown(struct link *link, struct link *requestor)
{
    struct list_head *pos, *n;

    l_debug("tearing down link [%08x] at 0x%p\n", link->id, link);

    if ( link->keepalive_id ) {
        g_source_remove(link->keepalive_id);
        link->keepalive_id = 0;
        ref_release(link);
    }

    /*
     * we should inform our neighbour that the link is going down so
     * that they don't bother to get to us anymore.  
     */
    if ( link != requestor ) {
        struct packet *pkt = pkt_getref_alloc(0, 0);

        if ( pkt ) {
            outgoing_link(pkt, link, LINK_DESTROY, 0);
            ref_release(pkt);
        }
    }
    /* 
     * we also tear down all the tunnels associated with this link.  the tunnel
     * teardown won't send tunnel destruction announcements over this link
     * 'cause we're claiming it initiated the teardown.
     */
    list_for_each_safe(pos, n, &link->tun_fwd_head) {
        struct tunnel *tunnel = list_entry(pos, struct tunnel, fwd_iterator); 

        tunnel_teardown(tunnel, link);
    }
    list_for_each_safe(pos, n, &link->tun_rev_head) {
        struct tunnel *tunnel = list_entry(pos, struct tunnel, rev_iterator); 

        tunnel_teardown(tunnel, link);
    }

    if ( link->up ) {
        link->up = 0;
        topo_link_down(link);
    }

    if ( ! list_empty(&link->sk_entry) )
        list_del(&link->sk_entry);

    g_hash_table_remove(link->sk->link_id_hash, &link->id);
    g_hash_table_remove(link->sk->link_nei_hash, &link->sin);
}

struct link *
link_getref_alloc(struct socket *sk)
{
    struct link *link;

    if ( (link=g_new0(struct link, 1)) == NULL)
        return NULL;

    link->neg_fwd_hash = g_hash_table_new(neghdr_hash, neghdr_equal);
    link->local_id_hash = g_hash_table_new(u32_hash, u32_equal);

    if ( !link->neg_fwd_hash || !link->local_id_hash )
        goto error;

    link->sk = sk;
    link->neighbour = 1; /* XXX doesn't belong here.. */
    seq_init(&link->seq);

    INIT_LIST_HEAD(&link->tun_fwd_head);
    INIT_LIST_HEAD(&link->tun_rev_head);
    INIT_LIST_HEAD(&link->sk_entry);

    REF_INIT(link, link_destroy);

    l_debug("allocated link %8x at %p\n", link->id, link);
    
    return link;

error:
    if ( link->local_id_hash )
        g_hash_table_destroy(link->local_id_hash);
    if ( link->neg_fwd_hash )
        g_hash_table_destroy(link->neg_fwd_hash);
    free(link);

    return NULL;
}

static void
link_id_assign(struct link *link)
{
    if ( link->id )
        return;

    link->id = get_new_id(link->sk->link_id_hash);
    g_hash_table_insert(link->sk->link_id_hash, &link->id, ref_get(link));
    list_add(&link->sk_entry, &link->sk->links_head);
    l_debug("assigned id 0x%8x to link %p\n", link->id, link);
}

struct link *
link_id_getref(struct socket *sk, guint32 id)
{
    struct link *link;

    if ((link = g_hash_table_lookup(sk->link_id_hash, &id)) == NULL)
        return NULL;

    return ref_get(link);
}

struct link *
link_nei_getref(struct socket *sk, struct sockaddr_in *sin)
{
    struct link *link;

    if ((link = g_hash_table_lookup(sk->link_nei_hash, sin)) == NULL)
        return NULL;

    return ref_get(link);
}

static gboolean
link_set_keepalive(gpointer data)
{
    struct link *link = data;
    time_t idle = time(NULL) - link->activity;
    /* XXX config in seconds.. */
    time_t link_keepalive_threshhold = 300;
    time_t link_keepalive_interval = 30;
    time_t link_keepalive_timeout = 600;

    l_debug("link %08x idle %ld seconds\n", link->id, idle);

    if ( idle >= link_keepalive_timeout ) {
        l_debug("keepalive removing link\n");
        link->keepalive_id = 0;
        link_teardown(link, NULL);
    } else 
    if ( idle >= link_keepalive_threshhold ) {
        /*
         * we've been idle for too long, start sending keepalive
         * probes..
         */
        link_send_message(link, LINK_ECHO_REQUEST);
        link->keepalive_id = g_timeout_add(link_keepalive_interval * 1000, 
                                            link_set_keepalive, 
                                            ref_get(link));
    } else {
        /*
         * we've been active within the keepalive threshhold, set our next
         * timeout to when we'd hit the idle threshhold if there was no
         * more activity
         */
        link->keepalive_id = 
            g_timeout_add((link_keepalive_threshhold - idle) * 1000, 
                                        link_set_keepalive, ref_get(link));
    }

    ref_release(link);
    return FALSE;
}

/*
 * XXX 
 *     this needs to be more careful.. it needs to work if its called
 *     more than once for a given link.  (neg retransmissions)
 */
static void
link_finalize(struct link *link, struct secret *sec)
{
    if ( link->session == NULL ) {
        link->session = sec;
        /* XXX check for errors */
        link->left = keys_setup(sec, CIPHER_BF, link->neg_initiator, 
                "HERE COMES THE CASBAH!");
        link->right = keys_setup(sec, CIPHER_BF, link->neg_initiator, 
                "CAN YOU ROCK IT!?");
    } else 
        secret_free(sec);

    if ( link->neighbour ) 
            g_hash_table_insert(link->sk->link_nei_hash, 
                                        &link->sin, ref_get(link));
    
    link->activity = time(NULL);
    link_set_keepalive(ref_get(link));
}

void
link_up(struct link *link)
{
    link->up = 1;
    topo_link_up(link->sk, link);
}

/*
 * just send out a quick message that doesn't have any payload..
 */
void 
link_send_message(struct link *link, guint8 type)
{
    struct packet *pkt = pkt_getref_alloc(0, 0);

    if ( pkt ) {
        outgoing_link(pkt, link, type, 0);
        ref_release(pkt);
    }
}

/*
 * ------------------------------------------------------------
 * interface into the rekeying bits..
 * ------------------------------------------------------------
 */
void link_rekey_action(struct link *link, int actions) 
{
    if ( actions & REK_DERIVE ) {
        l_debug("deriving new secondary keys\n");
        keys_derive(link->left);
        keys_derive(link->right);
        link->valid_new = 1;
    }
    if ( actions & REK_REQUEST ) {
        l_debug("sending a rekey request\n");
        link_send_message(link, LINK_REKEY_REQUEST);
    }
    if ( actions & REK_FLIP ) {
        l_debug("flipping keys\n");
        SWAP(link->right->in, link->right->in_new);
        SWAP(link->right->out, link->right->out_new);
        SWAP(link->left->in, link->left->in_new);
        SWAP(link->left->out, link->left->out_new);
    }
    if ( actions & REK_DESTROY ) {
        l_debug("ignoring secondary keys\n");
        link->valid_new = 0;
    }
    if ( actions & REK_SEQRESET ) {
        l_debug("reseting sequence number\n");
        seq_init(&link->seq);
    }
    if ( actions & REK_SEND ) {
        l_debug("sending something with primary keys\n");
        /*
         * using echo request for this is kind of funny.. it causes
         * the other side to reply when they tend to also be sending
         * their own requests.. it adds a bit of redundancy.
         */
        link_send_message(link, LINK_ECHO_REQUEST);
    }
}

/*
 * ------------------------------------------------------------
 * our interface into the negotiation code
 * ------------------------------------------------------------
 */

static void
send_packet(struct packet *pkt, void *state)
{
    struct link *link = state;

    socket_sendto(link->sk, &link->sin, pkt);
}

static void
append_options(struct packet *pkt, void *state)
{
    return;
}

static guint16
parse_confreq(struct packet *pkt, struct packet *out,
        void *state, struct secret *sec)
{
    struct link *link = state;

    link_id_assign(link);
    link_finalize(link, sec);

    return NEG_ERR_OK;
}

/*
 * this completes a negotiation and brings a real link
 * into being.
 */
static void
got_confack(struct packet *pkt, void *state, struct secret *sec)
{
    struct link *link = state;

    /*
     * XXX should be checking the result and unlink()ing if it failed
     */

    link_finalize(link, sec);
    link_up(link);
}

static void
release(void *state)
{
    struct link *link = state;
    ref_release(link);
}

static void
get_ref(void *state)
{
    struct link *link = state;
    ref_get(link);
}

struct neg_callbacks link_neg_cb = {
    append_options:     append_options,
    parse_confreq:      parse_confreq,
    got_confack:        got_confack,
    send_packet:        send_packet,
    get_ref:            get_ref,
    release:            release,
};

void
link_process_neg(struct socket *sk, struct packet *pkt, 
        struct sockaddr_in *from)
{
    struct link *link;
    struct addr_junk {
        guint32 ip;
        guint16 port;
    } __attribute__ ((packed)) addr_junk;

    /*
     * we create some simple state so that the async completion knows
     * who to send the response to...
     */
    if ( (link = link_getref_alloc(sk))  == NULL) 
        return;

    memcpy(&link->sin, from, sizeof(struct sockaddr_in));

    addr_junk.ip = from->sin_addr.s_addr;
    addr_junk.port = from->sin_port;

    neg_incoming(pkt, &link_neg_cb, link, &addr_junk, sizeof(addr_junk));

    ref_release(link);

    return;
}

void 
link_prepare(struct socket *sk, struct sockaddr_in *sin)
{
    struct link *link;

    link = link_getref_alloc(sk);
    if (link == NULL) {
        l_critical("couldn't allocate memory for link\n");
        return;
    }

    link_id_assign(link);
 
    memcpy(&link->sin, sin, sizeof(link->sin));
    link->neg_initiator = 1;

    ref_release(link);
}

static void 
link_neg_start(gpointer key, gpointer data, gpointer user_data)
{
    struct link *link = data;

    if ( link->neg_initiator )
        neg_initiate(&link_neg_cb, link);
}

void
link_hash_delete(gpointer data)
{
    struct link *link = data;
    ref_release(link);
}

static void
link_shutdown(gpointer data)
{
    struct daemon *dae = data;
    GList *list;
    struct list_head *pos, *n;
    struct link *link;

    for ( list = dae->socket_list ; list != NULL ; list = list->next) {
        struct socket *sk = list->data;

        list_for_each_safe(pos, n, &sk->links_head) {
            link = list_entry(pos, struct link, sk_entry);

            link_teardown(link, NULL);
        }
    }
}

int
link_init(struct daemon *dae)
{
    GList *list;

    if ( hook_register(&dae->shutdown_hooks, link_shutdown, dae) )
        return 0;

    for ( list = dae->socket_list ; list != NULL ; list = list->next) {
        struct socket *sk = list->data;
        g_hash_table_foreach(sk->link_id_hash, link_neg_start, NULL);
    }

    return 1;
}
