/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *    (c) Copyright 2001 Phil Schwan <phil@off.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/errno.h>
#include <signal.h>

#include <autoconf.h>
#ifdef USE_MTRACE
# include <mcheck.h>
#endif

#include "log.h"
#include "main.h"
#include "config.h"
#include "udp.h"
#include "pkt.h"
#include "neg.h"
#include "random.h"
#include "defaults.h"
#include "plugin.h"
#include "secret.h"
#include "keys.h"
#include "auth.h"
#include "tunnel.h"

static void
usage(char *name)
{
    fprintf(stderr, "Usage: %s [-c file] [-h] [-v]\n", name);
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "  -h        : list command line options (this message)\n");
    fprintf(stderr, "  -c config : specify an alternate config file\n");
    fprintf(stderr, "              (the default is %s)\n", CEBOLLA_CONFIG);
    fprintf(stderr, "  -v        : show version information\n");
    fprintf(stderr, "  -d        : enable debugging messages\n");
    fprintf(stderr, "  -v path   : chroot to path before parsing config\n");
    fprintf(stderr, "  -u uid    : run daemon as this numerical user id\n");
    fprintf(stderr, "  -g gid    : run daemon as this numerical group id\n");
    fprintf(stderr, "  -f        : run the daemon in the foreground\n");
}

static void
version(void)
{
    fprintf(stderr, "cebolla version %s, built on %s\n",
            CEBOLLA_VERSION, BUILD_DATE);
    fprintf(stderr, "http://www.zabbo.net/cebolla/\n");
}

int 
daemonize(char *root, uid_t uid, gid_t gid, int foreground)
{
    gid_t cur_gid = getgid();
    uid_t cur_uid = getuid();

    if ( root && chroot(root) ) {
        fprintf(stderr, "could not chroot(%s): %s\n", root, strerror(errno));
        return -1;
    }

    if ( gid && (gid != cur_gid) && setgid(gid) ) {
        fprintf(stderr, "could not setgid(%d): %s\n", gid, strerror(errno));
        return -1;
    }

    if ( uid && (uid != cur_uid) && setuid(uid) ) {
        fprintf(stderr, "could not setuid(%d): %s\n", uid, strerror(errno));
        return -1;
    }

    if ( foreground )
        return 0;

    switch( fork() ) {
        case -1:
            perror("could not fork");
            return -1;

        default:
            /* parent */
            exit(0);

        case 0:
            /* child */
            break;
    }

    /* 
     * should be doing something about ttys and stuff, right?  
     * need to look at stevens again :)
     */
    return 0;
}

/*
 * UNIX makes me want to cry.
 */

static sig_atomic_t should_exit;

static void die_handler(int signum)
{
    should_exit = 1;
    signal(signum, die_handler);
}    

static gboolean
check_signals(gpointer data)
{
    struct daemon *dae = data;

    if ( should_exit )  {
        g_main_loop_quit(dae->mainloop);
        return FALSE;
    }

    return TRUE;
}

int 
main(int argc, char **argv) 
{
    struct daemon d, *dae = &d;
    char *configfile = CEBOLLA_CONFIG;
    int c;
    char *root = NULL;
    uid_t uid = 0;
    gid_t gid = 0;
    int foreground = 1;

#ifdef USE_MTRACE
    mtrace();
#endif

#ifdef USE_VULNERABILITY
    if ( getenv("CEBOLLA_VULNERABLE_DO_NOT_USE") == NULL ) {
        fprintf(stderr, 
                "This version of cebolla was compiled with \n"
                "--enable-vulnerability which is a debugging hook that\n"
                "sends all traffic over the wire in the clear.  You should\n"
                "recompile cebolla without this option before using it.\n");
        exit(1);
    }
#endif

    memset(dae, 0, sizeof(struct daemon));

    while ( (c = getopt(argc, argv, "dfu:g:r:c:hv")) != -1) {
        extern char *optarg;
        unsigned long ul;

        switch(c) {
            case 'g':
                ul = strtoul(optarg, NULL, 10);
                if ( ul == LONG_MAX ) {
                    fprintf(stderr, "%s is not a valid numerical gid\n", 
                            optarg);
                    exit(1);
                }
                gid = ul;
                break;
            case 'u':
                ul = strtoul(optarg, NULL, 10);
                if ( ul == LONG_MAX ) {
                    fprintf(stderr, "%s is not a valid numerical uid\n", 
                            optarg);
                    exit(1);
                }
                uid = ul;
                break;
            case 'r':
                root = optarg;
                break;
            case 'd':
                log_enable_level(G_LOG_LEVEL_DEBUG);
                break;
            case 'f':
                foreground = 1;
                break;
            case 'c':
                configfile = optarg;
                break;
            case 'v':
                version();
                exit(0);
            case 'h':
            case '?':
                usage(argv[0]);
                exit(1);
        };
    };

    if ( daemonize(root, uid, gid, foreground) ) 
        exit(1);

    g_hook_list_init(&dae->shutdown_hooks, 256); /* arbitrary big number */

    dae->mainloop = g_main_new(0);
    if (dae->mainloop == NULL) {
        fprintf(stderr, "couldn't allocate main loop\n");
        exit(1);
    }

    signal(SIGINT, die_handler);
    signal(SIGTERM, die_handler);
    g_timeout_add(1000, check_signals, dae);

    /*
     * the ordering of these is important, some of the later init_ 
     * functions depend on previous ones.  
     */

    if (    logging_init(dae)               &&
            random_init()                   &&
            config_parse(dae, configfile)   &&
            logging_init_with_config(dae)   &&
            tunnel_init()                   &&
            cmd_init(dae)                   &&
            auth_init(dae)                  &&
            secret_init()                   &&
            plugin_init(dae)                &&
            socket_init(dae)                &&
            keys_init()                     &&
            neg_init(dae)                   &&
            link_init(dae)                          ) {

                g_main_run(dae->mainloop);
    }

    g_hook_list_invoke(&dae->shutdown_hooks, FALSE);

    return 0;
}
