/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *    (c) Copyright 2001 Phil Schwan <phil@off.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>

#include <glib.h>

#include "neg.h"
#include "log.h"
#include "random.h"
#include "wire.h"
#include "debug.h"
#include "util.h"
#include "secret.h"
#include "digest.h"
#include "config.h"
#include "auth.h"

/*
 * the initiator maintains this state for negotiations.  since the 
 * code paths use it heavily, the responder just fills in a static
 * one of these for use by its code paths.. its regenerated for
 * each incoming packet.
 */

struct negotiation {
    /*
     * ->head always points to the w_neg_hdr
     */
    struct packet *pkt;
    struct w_neg_hdr nhdr;

    struct neg_callbacks *cb;
    void *state;

    gint timer_id;
    int retries_left;

    struct secret *sec;
    struct auth_instance *ainst;
    int retrans_timer;
    refcount_t __refcount;
    void (*__destroy)(struct negotiation *neg);
};

static struct digest *neg_dig;
static char id_secret[2][MAX_DIGEST_BYTES];
int id_secret_current = 0;
int recalc_timer;

/*
 * used to store initiator negotiation state..
 */
GHashTable *neg_hash;

static void
neg_destroy(struct negotiation *neg)
{
    l_debug("freeing neg: %p\n", neg);

    neg->cb->release(neg->state);

    if ( neg->pkt ) 
        ref_release(neg->pkt);

    if ( neg->retrans_timer ) 
        g_source_remove(neg->retrans_timer);

    if (neg->sec != NULL) 
        secret_free(neg->sec);

    if (neg->ainst != NULL)
        auth_instance_free(neg->ainst);

    free(neg);
}

static struct negotiation *
neg_getref_alloc(struct neg_callbacks *cb, void *state)
{
    struct negotiation *neg;

    neg = g_new0(struct negotiation, 1);
    if(neg == NULL) 
        return NULL;

    l_debug("allocated neg %p\n", neg);

    REF_INIT(neg, neg_destroy);

    neg->cb = cb;
    neg->state = state;

    neg->cb->get_ref(neg->state);

    return neg;
}

static struct negotiation *
neg_id_getref(struct w_neg_hdr *nhdr)
{
    struct negotiation *neg;

    neg = g_hash_table_lookup(neg_hash, nhdr->initiator_id);
    if(neg == NULL) {
        l_debug("couldn't find neg state\n");
        return NULL;
    }

    return ref_get(neg);
}

static void
neg_teardown(struct negotiation *neg)
{
    g_hash_table_remove(neg_hash, neg->nhdr.initiator_id);

    ref_release(neg);
}


/*
 * used by callers to associate their state with an incoming
 * packet before they hand the packet to the neg core.
 */
void *
neg_get_state_ref(struct packet *pkt)
{
    struct w_neg_hdr nhdr;
    struct negotiation *neg;
    void *state;

    if ( pkt->len < sizeof(struct w_neg_hdr) ) 
        return NULL;

    memcpy(&nhdr, pkt->head, sizeof(nhdr));

    if ( (neg = neg_id_getref(&nhdr)) == NULL )
        return NULL;

    state = neg->state;
    neg->cb->get_ref(state);

    ref_release(neg);

    return state;
}

static void 
generate_resp_id(struct w_neg_hdr *nhdr, void *addr, int addr_len)
{
    digest_update(neg_dig, addr, addr_len);
    digest_update(neg_dig, id_secret[id_secret_current], MAX_DIGEST_BYTES);
    digest_complete(neg_dig, nhdr->responder_id, NEG_ID_BYTES);
}

static int
valid_resp_id(struct w_neg_hdr *nhdr, void *addr, int addr_len)
{
    struct w_neg_hdr nh;
    int valid;

    generate_resp_id(&nh, addr, addr_len);
    valid = (! memcmp(nh.responder_id, nhdr->responder_id, NEG_ID_BYTES));

    if ( !valid ) {
        /* see if it used the old secret */
        id_secret_current ^= 1;
        generate_resp_id(&nh, addr, addr_len);
        id_secret_current ^= 1;
        valid = (! memcmp(nh.responder_id, nhdr->responder_id, NEG_ID_BYTES));
    }

    return valid;
}

static void
neg_send(struct negotiation *neg, struct packet *pkt, 
        guint8 type, guint16 error)
{
    pkt_pull_head(pkt, sizeof(struct w_neg_hdr));

    /* temporarily use the neg->hdr.type/id fields.. */
    neg->nhdr.type = type;
    neg->nhdr.error = htons(error); 
    memcpy(pkt->head, &neg->nhdr, sizeof(struct w_neg_hdr));

    l_debug("sending type %d error %d\n", type, error);

    /* 
     * dont send partial crap in error cases
     */
    if ( error ) 
        pkt_pull_tail(pkt, pkt->len - sizeof(struct w_neg_hdr));

    neg->cb->send_packet(pkt, neg->state);
}

static gboolean
neg_retrans(gpointer data)
{
    struct negotiation *neg = data;
    neg->cb->send_packet(neg->pkt, neg->state);

    return TRUE;
}

/*
 * ------------------------------------------------------------
 * lots of helpers for building and disecting the neg packets..
 * ------------------------------------------------------------
 */

/*
 * pull the secret negotiation data off the front of the packet
 * and process it, storing the result in neg for later use..
 */
static guint16
pluck_sec(struct negotiation *neg, struct packet *pkt)
{
    struct w_blob blob;

    if ( ! pkt_get_head_blob(pkt, &blob)) 
        return NEG_ERR_BAD_SEC;

    /*
     * XXX again, this probably needs to be thought about a little more..
     */
    if ( neg->sec == NULL )  {
        if ((neg->sec = secret_parse_ann(pkt, blob.type)) == NULL)
            return NEG_ERR_BAD_SEC;
    }


    pkt_push_head(pkt, blob.len);

    return NEG_ERR_OK;
}

/*
 * pull the authentication data off the front of the packet
 * and process it, storing the result in neg for later use..
 */
static guint16
pluck_auth(struct negotiation *neg, struct packet *pkt)
{
    struct w_blob blob;

    if (! pkt_get_head_blob(pkt, &blob))  
        return NEG_ERR_BAD_AUTH;

    /* XXX this can't be right..  
     */
    if ( neg->ainst == NULL ) {
        neg->ainst = auth_instance_new(blob.type);

        if (neg->ainst == NULL) 
            return NEG_ERR_BAD_AUTH;

        ref_get(pkt); /* XXX */
        neg->ainst->data = pkt->head;
        neg->ainst->len = blob.len;
    }

    pkt_push_head(pkt, blob.len);

    return NEG_ERR_OK;
}

/*
 * sign the full contents of a packet and append the signature..
 */
static guint16 
sign_pkt(struct packet *pkt, struct auth_instance *ainst)
{
    struct w_blob *bl;
    struct packet *clone;

    clone = pkt_clone(pkt);
    if ( clone == NULL )
        return NEG_ERR_TEMP_FAILURE;

    /* Sign the payload */
    bl = pkt_append_blob_start(pkt, B_AUTH_SIGNATURE);
    ainst->at->sign(ainst, pkt, clone->head, clone->len);
    pkt_append_blob_end(pkt, bl);

    ref_release(clone);

    return NEG_ERR_OK;
}

/*
 * verify that the signature at the end of the packet was generated using
 * the rest of the packet's payload.  we pull the signature off the end
 * of the pkt so that others don't have to worry about it
 */
static guint16 
pluck_verify_sig(struct packet *pkt, struct auth_instance *ainst)
{
    struct packet *clone;
    struct w_blob blob;
    guint16 error;

    clone = pkt_clone(pkt);
    if ( clone == NULL )
        return NEG_ERR_TEMP_FAILURE;

    /*
     * walk our private clone until we find the signature..
     */
    for ( ; pkt_get_head_blob(clone, &blob) ;  
                                     pkt_push_head(clone, blob.len) ) {

        l_debug("saw blob %d\n", blob.type);

        if ( blob.type != B_AUTH_SIGNATURE)
            continue;
        /*
         * trim off the signature data from the real packet and verify
         */
        pkt_pull_tail(pkt, blob.len + sizeof(blob));

        pkt_pull_head(clone, sizeof(blob));
        if (ainst->at->verify(ainst, pkt->head, pkt->len,
                              clone) == 0) {

            l_warn("invalid signature in NEG_REPLY\n");
            error = NEG_ERR_INVALID_SIGNATURE;
            goto out;
        }

        error = NEG_ERR_OK;
        goto out;
    }

    error = NEG_ERR_NO_SIGNATURE;

out:
    ref_release(clone);

    return error;
}
       
/*
 * we've been sent a negotiation request, we need to quickly generate
 * a response.  we don't want to build up any state doing this, and it
 * should be a very lightweight operation
 */
static guint16 
handle_neg_init(struct negotiation *neg, struct packet *pkt, 
    struct packet *out)
{
    int auth_type, sec_type;
    struct w_blob blob, *bl;

    /*
     * decide which auth we like best..
     */
    if (pkt_get_head_blob(pkt, &blob) && blob.type == B_AUTH_LIST) {

        auth_type = auth_choose((guint16 *)pkt->head,
                                blob.len / sizeof(guint16));
        pkt_push_head(pkt, blob.len);

    } else {
        l_warn("badly formed packet in NEG_INIT (missing or invalid "
               "AUTH_LIST blob\n");
        return NEG_ERR_BAD_AUTH_LIST; /* XXX leaking neg->ainst */
    }

    /*
     * .. and append our half to the packet.
     */
    neg->ainst = auth_instance_new(auth_type);
    if ( neg->ainst == NULL )
        return NEG_ERR_BAD_AUTH;

    bl = pkt_append_blob_start(out, auth_type);
    neg->ainst->at->append_data(neg->ainst, out);
    pkt_append_blob_end(out, bl);

    /*
     * now find which sec neg system we prefer..
     */
    if (pkt_get_head_blob(pkt, &blob) && blob.type == B_SEC_LIST) {

        sec_type = secret_choose((guint16 *)pkt->head,
                                 blob.len / sizeof(guint16));
        pkt_push_head(pkt, blob.len);

    } else {
        l_warn("badly formed packet in NEG_INIT (missing or invalid "
               "SEC_LIST blob\n");
        return NEG_ERR_BAD_SEC_LIST;
    }

    if (sec_type != B_SEC_DH) {
        l_warn("XXX temporarily forcing B_SEC_DH\n");
        sec_type = B_SEC_DH;
    }

    /*
     * and append our half to the outgoing packet
     */
    bl = pkt_append_blob_start(out, sec_type);
    secret_append_ann(out, sec_type);
    pkt_append_blob_end(out, bl);

    /*
     * XXX we should be caching signed copies of the auth/sec data to
     * be sent out, doing a full signature per packet here isn't quite
     * nice.
     */
    return sign_pkt(out, neg->ainst);
}

/*
 * we've been sent a reply.  we collect the auth data that was sent to
 * us and submit a config request..
 */
static guint16 
handle_neg_reply(struct negotiation *neg, struct packet *pkt, 
        struct packet *out)
{
    struct w_blob *bl;
    guint16 error;
    struct packet *clone;

    clone = pkt_clone(pkt);
    if ( clone == NULL )
        return NEG_ERR_TEMP_FAILURE;

    error = pluck_auth(neg, pkt);
    if ( error )  {
        ref_release(clone);
        return error;
    }

    error = pluck_verify_sig(clone, neg->ainst);
    ref_release(clone);
    if ( error ) 
        return error;

    error = pluck_sec(neg, pkt);
    if ( error ) 
        return error;

    /*
     * append our half of the secret negotiation and our authentication
     * data, based on the types that we found in the packet.
     */
    bl = pkt_append_blob_start(out, B_SEC_DH); /* XXX neg->sec, somehow */
    secret_append_ann(out, B_SEC_DH);
    pkt_append_blob_end(out, bl);

    bl = pkt_append_blob_start(out, neg->ainst->at->auth_id);
    neg->ainst->at->append_data(neg->ainst, out);
    pkt_append_blob_end(out, bl);

    neg->cb->append_options(out, neg->state);

    /* XXX should be encrypting auth and options */

    return sign_pkt(out, neg->ainst);
}            

static guint16 
handle_neg_confreq(struct negotiation *neg, struct packet *pkt, 
        struct packet *out)
{
    guint16 error;
    struct packet *clone;

    /*
     * the signature needs to see a snapshot of the whole packet
     */
    clone = pkt_clone(pkt);
    if ( clone == NULL )
        return NEG_ERR_TEMP_FAILURE;

    error = pluck_sec(neg, pkt);
    if ( error )  {
        ref_release(clone);
        return error;
    }

    /* XXX should be decrypting the auth/signature */

    error = pluck_auth(neg, pkt);
    if ( error )  {
        ref_release(clone);
        return error;
    }

    error = pluck_verify_sig(clone, neg->ainst);
    ref_release(clone);
    if ( error ) 
        return error;

    error = neg->cb->parse_confreq(pkt, out, neg->state, neg->sec);
    if ( error )
        return error;

    neg->sec = NULL;

    error = sign_pkt(out, neg->ainst);
    if ( error ) 
        return error;

    return NEG_ERR_OK;
}

static guint16 
handle_neg_confack(struct negotiation *neg, struct packet *pkt)
{
    guint16 error;

    if ( neg->ainst == NULL ) {
        l_warn("got a confack with no previous auth inst\n");
        return NEG_ERR_BAD_AUTH;
    }

    error = pluck_verify_sig(pkt, neg->ainst);
    if ( error ) 
        return error;

    /* XXX decrypt option list */

    neg->cb->got_confack(pkt, neg->state, neg->sec);
    neg->sec = NULL;

    return NEG_ERR_OK;
}

static void
initiator_incoming(struct packet *pkt, struct w_neg_hdr *nhdr,
        struct neg_callbacks *cb)
        
{
    struct negotiation *neg;
    guint16 error = 0;

    if ( (neg = neg_id_getref(nhdr)) == NULL )
        return;

    pkt_push_head(pkt, sizeof(struct w_neg_hdr));

    /*
     * reset the contents of our long lived packet that stores
     * the output of this stage in the negotiation for later
     * retransmission..
     */
    pkt_push_head(neg->pkt, sizeof(struct w_neg_hdr));
    pkt_pull_tail(neg->pkt, neg->pkt->len);

    switch(nhdr->type) {
        case NEG_REPLY:
            /*
             * save the responder id that was given to us 
             */
            memcpy(&neg->nhdr.responder_id, nhdr->responder_id, NEG_ID_BYTES);

            error = handle_neg_reply(neg, pkt, neg->pkt);
            if ( error ) 
                neg_teardown(neg);
            else {
                neg_send(neg, neg->pkt, NEG_CONFREQ, 0);
                g_source_remove(neg->retrans_timer);
                neg->retrans_timer = g_timeout_add(1000, neg_retrans, neg);
            }

            break;

        case NEG_CONFACK:

            handle_neg_confack(neg, pkt);
            neg_teardown(neg);
            break;
    }
    ref_release(neg);
}
static void
responder_incoming(struct packet *pkt, struct w_neg_hdr *nhdr,
        struct neg_callbacks *cb, void *state,
        void *addr, int addr_len)
{
    struct negotiation *neg;
    guint16 error;
    struct packet *out;
    /*
     * the stateless resonder doesn't have state for the ngotiation
     * already, but the neg processing code paths will want a negotiation
     * struct to sttach some of their state too.  we allocate this hear,
     * build it up just enough, and tear it down again :/
     */
    if ( (neg = neg_getref_alloc(cb, state)) == NULL)  
        return;

    /* 
     * XXX need a way to predict the payload size, this is silly.
     */
    if ( (out = pkt_getref_alloc(sizeof(*nhdr), 4096)) == NULL )
        goto release_neg;

    memcpy(neg->nhdr.initiator_id, nhdr->initiator_id, NEG_ID_BYTES);

    switch(nhdr->type) {

      case NEG_INIT:
            /*
             * generate a new responder id, unique to this negotiation.
             */
            generate_resp_id(&neg->nhdr, addr, addr_len);

            pkt_push_head(pkt, sizeof(struct w_neg_hdr));

            error = handle_neg_init(neg, pkt, out);

            neg_send(neg, out, NEG_REPLY, error);

            break;

      case NEG_CONFREQ:

            /*
             * we don't want to expend too much energy on people who
             * aren't Doing It Right.
             */
            if ( ! valid_resp_id(nhdr, addr, addr_len)) {
                l_debug("invalid responder id\n");
                break;
            }

            memcpy(&neg->nhdr.responder_id, nhdr->responder_id, NEG_ID_BYTES);

            pkt_push_head(pkt, sizeof(struct w_neg_hdr));

            error = handle_neg_confreq(neg, pkt, out);

            neg_send(neg, out, NEG_CONFACK, error);

            break;
    }

    ref_release(out);

release_neg:
    ref_release(neg);
}

void 
neg_incoming(struct packet *pkt, struct neg_callbacks *cb, void *state,
        void *addr, int addr_len)
{
    struct w_neg_hdr nhdr;

    /*
     * not interested in obviously bogus packets..
     */
    if ( pkt->len >= sizeof(struct w_neg_hdr) ) {

        memcpy(&nhdr, pkt->head, sizeof(nhdr));

        /*
         * XXX we should be noticing the error field in the incoming
         * packet and reacting based on a cb->error_rx(), or something.
         */
        if ( ntohs(nhdr.error) )
            l_debug("got error %d\n", ntohs(nhdr.error));

        switch(nhdr.type) {
            case NEG_INIT:
            case NEG_CONFREQ:
                  responder_incoming(pkt, &nhdr, cb, state, addr, addr_len);
                  break;
            case NEG_REPLY:
            case NEG_CONFACK:
                  initiator_incoming(pkt, &nhdr, cb);
        };

    }
}

void 
neg_initiate(struct neg_callbacks *cb, void *state)
{
    struct negotiation *neg;
    struct packet *pkt;
    struct w_blob *bl;

    if ( (neg = neg_getref_alloc(cb, state)) == NULL) 
        return;

    /* XXX need some way to guess size here */
    if ( (pkt = pkt_getref_alloc(sizeof(struct w_neg_hdr), 768)) == NULL)
        goto out;

    neg->pkt = pkt;
    get_random_bytes(neg->nhdr.initiator_id, NEG_ID_BYTES);
    g_hash_table_insert(neg_hash, neg->nhdr.initiator_id, ref_get(neg));

    bl = pkt_append_blob_start(pkt, B_AUTH_LIST);
    auth_append_list(pkt);
    pkt_append_blob_end(pkt, bl);

    bl = pkt_append_blob_start(pkt, B_SEC_LIST);
    secret_append_list(pkt);
    pkt_append_blob_end(pkt, bl);

    neg_send(neg, pkt, NEG_INIT, 0);
    neg->retrans_timer = g_timeout_add(1000, neg_retrans, neg);

out:
    ref_release(neg);
    return;
}

gboolean
recalc_id_secret(gpointer data)
{
    id_secret_current ^= 1;
    get_random_bytes(id_secret[id_secret_current], MAX_DIGEST_BYTES);
    l_debug("recalcing id secret\n");

    return TRUE;
}

int 
neg_init(struct daemon *dae)
{
    if ( (neg_hash = g_hash_table_new(neghdr_hash, neghdr_equal)) == NULL) {
            l_critical("couldn't initialize hash table\n");
            return 0;
    }

    if ( (neg_dig = digest_new(DIGEST_SHA1)) == NULL) {
            l_critical("no neg_dig?!  Very sad.\n");
            return 0;
    }

    /* get things going initially.. */
    recalc_id_secret(NULL);
    recalc_id_secret(NULL);
    recalc_timer = g_timeout_add(dae->config->neg_recalc_timeout * 1000,
                                 recalc_id_secret, NULL);

    return 1;
}
