/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4: 
 */

#ifndef _NEG_H_
#define _NEG_H_

#include <glib.h>

#include "pkt.h"
#include "secret.h"
#include "main.h"

/*
 * the shared negotiation code deals with negotating the cipher and 
 * secret for the resource.  it deals with maintaining the negotiation
 * state in the initiator, handles retransmissions, etc.  all the caller
 * needs to do is provide a way to send packets and deal with the higher
 * level options that are passed in the negotiation.
 */

struct neg_callbacks {

    void (*send_packet)(struct packet *pkt, void *state);

    /*
     * this should append the options an initiator wants to a confreq
     * packet.
     */
    void (*append_options)(struct packet *pkt, void *state);

    /*
     * parse the options in the confreq packet in pkt, append the options to
     * the confack packet in out.  this callback happens when the responder
     * builds up state, so it needs to know the negotiated secret.  returning
     * ok implies having taken ownership of the secret.  an error tells
     * neg that it should free the secret.
     */
    guint16 (*parse_confreq)(struct packet *pkt, struct packet *out, 
            void *state, struct secret *sec);

    /*
     * lets the initiator discover what options were decided upon.  this
     * is the last callback the initiator will receive (besides state 
     * destruction) so it should use this to build up state, or tear it
     * down if it decides the options were not good enough.   got_confack()
     * must deal with sec.. either grab its pointer or free it.
     */
    void (*got_confack)(struct packet *pkt, void *state, struct secret *sec);

    void (*get_ref)(void *state);
    void (*release)(void *state);
};

int neg_init(struct daemon *dae);

/*
 * called to initiate a negotiation.  it will call
 * cb->destroy() on failure, though, so you don't have to 
 * clean up the state argument if the fucntion fails.
 */
void neg_initiate(struct neg_callbacks *cb, void *state);

/*
 * process an incoming negotiation packet.  cb->destroy(state) will
 * be called when it is no longer needed, and that could be quite
 * a while for async negitiation steps.  This can mangle the packet
 * that is passed in..
 */
void neg_incoming(struct packet *pkt, struct neg_callbacks *cb,
                void *state, void *addr, int addr_len);

void * neg_get_state_ref(struct packet *pkt);

#endif

