/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <netinet/in.h>
#include <string.h>

#include "pkt.h"
#include "log.h"
#include "debug.h"
#include "util.h"

#define MIN_HEADROOM 64

#define pkt_panic(pkt, args...) {                                           \
    l_debug(args);                                                          \
    l_error("pkt dump: [st:%p (%d) h:%p (%d) t:%p (%d) e:%p len: %d]\n",    \
            pkt->start,                                                     \
            (char *)pkt->head - (char *)pkt->start,                         \
            pkt->head,                                                      \
            (char *)pkt->tail - (char *)pkt->head,                          \
            pkt->tail,                                                      \
            (char *)pkt->end - (char *)pkt->tail,                           \
            pkt->end,                                                       \
            pkt->len);                                                      \
    }

/*
 * the exported functions..
 */

void *
pkt_push_head(struct packet *pkt, unsigned int len)
{
    char *region;

    if (len > pkt->len) 
       pkt_panic(pkt, "pushed %d\n", len);

    region = pkt->head;

    pkt->head += len;
    pkt->len -= len;

    return region;
}

void *
pkt_pull_head(struct packet *pkt, unsigned int len)
{
    if (((char *)pkt->head - len) < pkt->start) 
        pkt_panic(pkt, "pulled %d\n", len);

    pkt->head -= len;
    pkt->len += len;

    return pkt->head;
}

void *
pkt_push_tail(struct packet *pkt, unsigned int len)
{
    char *region;

    /*
     * this is why its rarely a good idea to directly dereference
     * ->tail.. someone may have it _right_ at the end.
     */
    if (((char *)pkt->tail + len) > pkt->end) 
        pkt_panic(pkt, "bad len %d\n", len);

    region = pkt->tail;
    pkt->tail += len;
    pkt->len += len;

    return region;
}

void *
pkt_pull_tail(struct packet *pkt, unsigned int len)
{
    if (len > pkt->len) 
        pkt_panic(pkt, "bad len %d\n", len);

    pkt->tail -= len;
    pkt->len -= len;

    return pkt->tail;
}

/*
 * the ever cute blob functions
 */

struct w_blob *
pkt_append_blob_start(struct packet *pkt, guint16 type)
{
    struct w_blob *blob;

    blob = pkt_push_tail(pkt, sizeof(struct w_blob));
    blob->type = htons(type);

    return blob;
}

void
pkt_append_blob_end(struct packet *pkt, struct w_blob *blob)
{
    guint16 len;

    len = ((char *)pkt->tail - (char *)blob) - sizeof(struct w_blob);
    l_debug("appended blob type %d of %d bytes\n", ntohs(blob->type), len);
    blob->len = htons(len);
}

void *
pkt_append_blob(struct packet *pkt, guint16 type, void *data, guint16 len)
{
    char *start, *region;
    struct w_blob blob;

    blob.len = htons(len);
    blob.type = htons(type);

    start = region = pkt_push_tail(pkt, sizeof(blob));
    memcpy(region, &blob, sizeof(blob));
    region = pkt_push_tail(pkt, len);
    memcpy(region, data, len);

    return start;
}

int pkt_get_head_blob(struct packet *pkt, struct w_blob *blob)
{
    if ( pkt->len < sizeof(struct w_blob) )
        return 0;

    hexdump(pkt->head, pkt->len, "at get head blob\n");

    memcpy(blob, pkt->head, sizeof(struct w_blob));
    blob->len = ntohs(blob->len);
    blob->type = ntohs(blob->type);

    l_debug("got type %d len %d\n", blob->type, blob->len);

    if ( (blob->len + sizeof(struct w_blob)) > pkt->len )
        return 0;

    pkt_push_head(pkt, sizeof(struct w_blob));

    return 1;
}

static void
pkt_destroy(struct packet *pkt)
{
    if ( pkt->is_clone )
        ref_release(pkt->parent);

    l_debug("freeing pkt: %p\n", pkt);

    free(pkt);
}

struct packet * 
pkt_getref_alloc(unsigned int headroom, unsigned int len)
{
    struct packet *pkt;

    headroom += MIN_HEADROOM;

    pkt = calloc(sizeof(struct packet) + len + headroom, 1);
    if ( pkt == NULL )
        return NULL;

    pkt->start = pkt->data;
    pkt->end = pkt->data + len + headroom;
    REF_INIT(pkt,  pkt_destroy);

    pkt->head = pkt->start + headroom;
    pkt->tail = pkt->head;
    pkt->len = 0;

    l_debug("allocted pkt: %p headroom: %d len: %d\n", pkt, headroom, len);

    return pkt;
}

struct packet *
pkt_clone(struct packet *pkt)
{
    struct packet *clone_pkt;

    clone_pkt = malloc(sizeof(struct packet));
    if ( clone_pkt == NULL )
        return NULL;

    memcpy(clone_pkt, pkt, sizeof (struct packet));

    REF_INIT(clone_pkt, pkt_destroy);
    clone_pkt->is_clone = 1;
    clone_pkt->parent = ref_get(pkt);

    return clone_pkt;
}

