/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2001 Phil Schwan <phil@off.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include "log.h"
#include "plugin.h"

static GList *plugin_list = NULL;

struct plugin_data *
plugin_load(struct config *config, const char *filename)
{
    struct plugin_data *data;
    int err;

    if (filename == NULL)
        l_error("filename == NULL\n");

    data = g_new0(struct plugin_data, 1);
    if (data == NULL) {
        l_warn("Unable to load plugin '%s': out of memory\n", filename);
        return NULL;
    }

    data->handle = g_module_open(filename, 0);
    if (data->handle == NULL) {
        l_warn("Unable to load plugin '%s': %s\n", filename, g_module_error());
        g_free(data);
        return NULL;
    }

    if (!g_module_symbol(data->handle, "init_plugin",
                         (gpointer *) &data->init_plugin)) {
        l_warn("Plugin '%s' lacks initialization function.\n", filename);
        g_module_close(data->handle);
        g_free(data);
        return NULL;
    }

    err = data->init_plugin(data, config);
    if (err < 0) {
        l_warn("%s::init_plugin returned error: %s\n",
               filename, g_strerror(-err));
        g_module_close(data->handle);
        g_free(data);
        return NULL;
    }

    plugin_list = g_list_prepend(plugin_list, data);
    return data;
}

void
plugin_unload(struct plugin_data *data)
{
    if (data == NULL)
        l_error("data == NULL\n");

    if (data->can_unload != NULL && !data->can_unload(data)) {
        l_warn("plugin_unload() called on plugin still in use (%s)\n",
            g_module_name(data->handle));
        return;
    }

    if (data->cleanup != NULL)
        data->cleanup(data);

    plugin_list = g_list_remove(plugin_list, data);
    if (!g_module_close(data->handle)) {
        l_warn("g_module_close(%s) failed: %s\n",
            g_module_name(data->handle), g_module_error());
        g_free(data);
    }
}

int
plugin_init(struct daemon *dae)
{
    DIR *d;
    struct dirent *e;

    if (!g_module_supported()) {
        l_warn("Not loading plugins; this system doesn't support them.\n");
        return 1;
    }

    if (global_config->plugin_directory == NULL) {
        l_warn("Not loading plugins; no plugin directory specified.\n");
        return 1;
    }

    if ((d = opendir(global_config->plugin_directory)) == NULL) {
        l_warn("Unable to open plugin directory '%s': %s\n",
               global_config->plugin_directory, strerror(errno));
        return 0;
    }

    while ((e = readdir(d)) != NULL) {
        char *p;

        if ((p = strrchr(e->d_name, '.')) && strcmp(p, ".so") == 0) {
            char *name = g_strconcat(global_config->plugin_directory,
                                     "/", e->d_name, NULL);
            plugin_load(dae->config, name);
            g_free(name);
        }
    }
    closedir(d);

    return 1;
}
