/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2001 Phil Schwan <phil@off.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <glib.h>

#define LOG_FLAGS LF_AUTH
#include "auth.h"
#include "config.h"
#include "log.h"
#include "pkt.h"
#include "plugin.h"
#include "debug.h"
#include "shell.h"
#include "util.h"

/*
 * Note that this plugin is basically just to show that it really is possible to
 * use the shell functions to write an auth plugin.  All of this functionality
 * is provided by the PGP auth plugin; the PGP auth plugin uses some code written
 * by Adam Back to avoid the seventy-five fork()s that it takes it use GPG.  If
 * you're using this plugin for anything more than testing, you're out of your
 * mind.
 */

/*
 * Before this plugin can be used, the administrator is expected to have created
 * a certificate (through some external, unspecified means, although we provide a
 * suggested setup with cebolla) and installed the keys in the GPG keyring.
 *
 * When this plugin is registered, the certificate (specified in the config file)
 * is loaded into the auth_type data field.  If this fails, we're finished before
 * we've even begun, so registration fails.
 *
 * We sign data with 'gpg --detach-sign --output -', which takes the payload from
 * stdin and writes the signed data to stdout.  Nothing could be simpler.
 *
 * Verification is significantly more trouble; we have to extract the certificate
 * from the payload handed to us, decouple the signature and non-signature data
 * in the certificate, verify that the signature is valid (thus verifying the
 * authenticity of the certificate), find the public key in the certificate,
 * install that key in GPG's public key ring, and then _finally_ verify that the
 * payload data was signed with the public key that we just installed.  It's a
 * good thing there's no libgpg so we have to fork 3 times; yes sir!  Good thing!
 */

static int
gpg_append_data(struct auth_instance *inst, struct packet *pkt)
{
    /* Append our certificate */
    void *buf = pkt_push_tail(pkt, inst->at->len);
    memcpy(buf, inst->at->data, inst->at->len);
    return 0;
}

static int
gpg_sign(struct auth_instance *inst, struct packet *pkt,
         unsigned char *data, int len)
{
    char * const argv[] = {"gpg", "--detach-sign", "--output", "-", NULL};
    int size = 512, retval = 0;
    void *buf;
    struct shell_handle *handle;

    if ((handle = shell_start("gpg", argv)) == NULL)
        return 0;

    /* Write the data to be signed */
    if (shell_write(handle, data, len) != len) {
        l_warn("shell_write failed.\n");
        return 0;
    }
    /* Close the pipe */
    shell_write(handle, NULL, 0);

    /* Read all of the data into the packet */
    buf = pkt_push_tail(pkt, size);
    while (1) {
        int tmp = shell_read_stdout(handle, buf, size);

        if (tmp < 0) {
            l_warn("shell_read failed: %d (%s)\n", errno, g_strerror(errno));
            break;
        }
        if (tmp == 0) {
            /* Finish reading */
            pkt_pull_tail(pkt, size - tmp);
            break;
        }
        size -= tmp;
        buf += tmp;
        retval += tmp;
        if (size < 0)
            l_error("buffer overrun?  what happened?\n");
        if (size == 0) {
            pkt_push_tail(pkt, 512);
            size += 512;
        }
    }
    shell_finish(handle);

    return retval;
}

/* Search 'size'-byte buffer 'buf' for 'string' */
static char *
memstr(char *buf, char *string, unsigned int size)
{
    int len = strlen(string);
    char *end = buf + size - len + 1, first;

    if (!len) {
        /* Don't abuse my function and I won't abuse you. */
        abort();
    }

    first = string[0];

    while (buf < end) {
        char *p = memchr(buf, first, end - buf);
        if (p == NULL)
            break;
        if (memcmp(p, string, len) == 0)
            return p;
        buf = p + 1;
    }

    return NULL;
}

/* Sheesh, I'd forgotten how much of a pain in the ass this was.   The first time
 * I wrote this--for Chainsaw--I never _imagined_ that I'd write it again.
 *
 * This function has to do a lot of stuff, described briefly above.  In more
 * detail:
 *
 * 1. Validate the cebolla certificate
 * 2. Find the signature within the certificate
 * 3. Verify the certificate's signature
 * 4. Find the public key within the certificate
 * 5. Import that public key into the local GPG key ring
 * 6. Verify the signature on the packet data, and see that the signing key ID
 *    matches the one that we just imported.
 *
 * FIXME:
 * - Need to be more careful with key IDs and ID strings
 * - error handling sucks; can leak tmpfiles.  do a quick cleanup.
 */
static int
gpg_verify(struct auth_instance *inst, unsigned char *data, int len,
           struct packet *pkt)
{
    struct w_blob sig_blob;
    char *sig_head, *key_start, *key_end, *cert_start, *cert_end, *p, *end;
    char tmpfile[15]="/tmp/cebXXXXXX", keyid[9] = "";
    char * const verify_argv[] = {"gpg", "--verify", tmpfile, "-", NULL};
    char * const import_argv[] = {"gpg", "--import", "-", NULL};
    char buf[4096], *b = buf;
    struct shell_handle *handle;
    int fd;

    end = inst->data + inst->len;
    if (memcmp(inst->data, "-----BEGIN CEBOLLA CERTIFICATE-----\n", 36) != 0) {
        l_debug("missing BEGIN CEBOLLA CERTIFICATE line\n");
        return 0;
    }
    cert_start = inst->data + 36;

    /* Hunt for the signature in this certificate (2 above) */
    p = memstr(cert_start, "-----BEGIN PGP SIGNATURE-----\n", inst->len - 36);
    if (p == NULL) {
        l_debug("missing BEGIN PGP SIGNATURE line\n");
        return 0;
    }
    cert_end = p;
    sig_head = p + 30;

    p = memstr(sig_head, "-----END PGP SIGNATURE-----\n", end - sig_head);
    if (p == NULL) {
        l_debug("missing END PGP SIGNATURE line\n");
        return 0;
    }

    /* Verify this signature.  This means a tmpfile.  HOORAY. */
    if ((fd = mkstemp(tmpfile)) < 0) {
        l_debug("mkstemp failed: %s\n", g_strerror(errno));
        return 0;
    }
    if (write(fd, sig_head, p - sig_head) < p - sig_head) {
        l_debug("write failed while writing signature into tmpfile: %s\n",
                g_strerror(errno));
        close(fd);
        return 0;
    }
    close(fd);

    /* Verify that the signature covering the certificate (a) is valid and
     * (b) was signed by the certificate authority (3 above) */
    if ((handle = shell_start("gpg", verify_argv)) == NULL) {
        l_debug("shell_start failed\n");
        return 0;
    }

    /* Write the certificate payload into the pipe */
    if (shell_write(handle, cert_start, cert_end - cert_start) !=
        cert_end - cert_start) {
        l_warn("shell_write failed.\n");
        return 0;
    }
    shell_write(handle, NULL, 0);

    b = buf;
    while (1) {
        int tmp = shell_read_stderr(handle, b, sizeof(buf) - (b - buf));

        if (tmp <= 0) {
            l_debug("couldn't find 'Good signature from CA' line (1)\n");
            l_debug("buf: \n%s\n", buf);
            shell_finish(handle);
            return 0;
        }

        b += tmp;
        /* FIXME: The CA's ID string and DSA key ID should be in the config
         * file, and verified here.  This is not good enough! */
        p = memstr(buf, "\ngpg: Good signature from \"cebolla certificate authority ", b - buf);
        if (p != NULL) {
            /* Success.  Move on. */
            break;
        }
        if (b - buf >= sizeof(buf)) {
            l_debug("couldn't find 'Good signature from CA' line (2)\n");
            l_debug("buf: \n%s\n", buf);
            shell_finish(handle);
            return 0;
        }
    }
    shell_finish(handle);
    unlink(tmpfile);

    /* Somewhere in that certificate is this entity's public key.  Find it.
     * (4 above) */
    p = memstr(cert_start, "-----BEGIN PGP PUBLIC KEY BLOCK-----\n",
               cert_end - cert_start);
    if (p == NULL) {
        l_debug("couldn't find BEGIN public key in cert\n");
        return 0;
    }
    key_start = p + 37;

    p = memstr(key_start, "-----END PGP PUBLIC KEY BLOCK-----",
               cert_end - key_start);
    if (p == NULL) {
        l_debug("couldn't find END public key in cert\n");
        return 0;
    }
    key_end = p - 1;

    /* Import the key. (5 above) */
    if ((handle = shell_start("gpg", import_argv)) == NULL) {
        l_debug("shell_start failed\n");
        return 0;
    }

    if (shell_write(handle, key_start, key_end - key_start) !=
        key_end - key_start) {
        l_warn("shell_write failed.\n");
        return 0;
    }
    shell_write(handle, NULL, 0);

    b = buf;
    while (1) {
        int tmp = shell_read_stderr(handle, b, sizeof(buf) - (b - buf));

        if (tmp <= 0) {
            l_debug("couldn't find 'key imported' line (1)\n");
            shell_finish(handle);
            return 0;
        }

        b += tmp;
        p = memstr(buf, ": public key imported\n", b - buf);
        if (p != NULL) {
            memcpy(keyid, p - 8, 8);
            break;
        }
        p = memstr(buf, ": not changed\n", b - buf);
        if (p != NULL) {
            memcpy(keyid, p - 8, 8);
            break;
        }
        if (b - buf > sizeof(buf)) {
            l_debug("couldn't find 'key imported' line (2)\n");
            shell_finish(handle);
            return 0;
        }
    }
    shell_finish(handle);

    /* Another blob should, in theory, contain the packet signature.  If not,
     * well, life's tough all over. */
    if (!pkt_get_head_blob(pkt, &sig_blob) ||
        sig_blob.type != B_AUTH_SIGNATURE) {
        l_warn("missing or invalid second auth blob\n");
        return 0;
    }

    memcpy(tmpfile, "/tmp/cebXXXXXX", sizeof(tmpfile));
    if ((fd = mkstemp(tmpfile)) < 0) {
        l_debug("mkstemp failed: %s\n", g_strerror(errno));
        return 0;
    }
    if (write(fd, pkt->head, sig_blob.len) < sig_blob.len) {
        l_debug("write failed while writing packet signature into tmpfile: %s\n",
                g_strerror(errno));
        close(fd);
        return 0;
    }
    close(fd);

    /* Verify that the signature covering the packet (a) is valid and
     * (b) was signed by the key in the certificate that we just validated. */
    if ((handle = shell_start("gpg", verify_argv)) == NULL) {
        l_debug("shell_start failed\n");
        return 0;
    }

    /* Write the certificate payload into the pipe */
    if (shell_write(handle, data, len) != len) {
        l_warn("shell_write failed.\n");
        return 0;
    }
    shell_write(handle, NULL, 0);

    b = buf;
    while (1) {
        char new_keyid[9] = "";
        int tmp = shell_read_stderr(handle, b, sizeof(buf) - (b - buf));

        if (tmp <= 0) {
            l_debug("couldn't find final 'Good signature' line (1)\n");
            shell_finish(handle);
            return 0;
        }
        b += tmp;
        if (b - buf <= 78 + 20)
            continue;

        p = memstr(buf, "\ngpg: Good signature from \"cebolla certificate authority ", b - buf);
        if (p == NULL)
            continue;

        if (memcmp(p - 8, keyid, 8) != 0) {
            memcpy(new_keyid, p - 8, 8);
            l_debug("KeyIDs differ: %s vs. %s\n", keyid, new_keyid);
            l_debug("buf: \n%s\n", buf);
            shell_finish(handle);
            return 0;
        } else {
            break;
        }

        if (b - buf > sizeof(buf)) {
            l_debug("couldn't find final 'Good signature' line (2)\n");
            shell_finish(handle);
            return 0;
        }
    }
    shell_finish(handle);
    unlink(tmpfile);

    /* Well, was that worth it?  No.  Probably not.  But there you go.*/

    return 1;
}

struct auth_type gpg_type = {
    name:        "GPG auth type",
    auth_id:     B_AUTH_GPG,
    start_inst:  NULL,
    stop_inst:   NULL,
    append_data: gpg_append_data,
    sign:        gpg_sign,
    verify:      gpg_verify,
    shutdown:    NULL
};

int
init_plugin(struct plugin_data *data, struct config *config)
{
    struct stat sb;
    int fd;

    if (config->pgp_certificate == NULL) {
        l_warn("You must specify a \"pgp_certificate\" in the configuration file before you can use the GPG auth plugin.  Plugin registration aborted.\n");
        return -1;
    }

    if (stat(config->pgp_certificate, &sb) < 0) {
        l_warn("Cannot find \"%s\": %s\n", config->pgp_certificate,
               g_strerror(errno));
        return -1;
    }

    if (sb.st_size > 64*1024) {
        l_warn("Your \"pgp_certificate\" is very large.  Is that actually a valid certificate?\n");
        return -1;
    }

    fd = open(config->pgp_certificate, O_RDONLY);
    if (fd < 0) {
        l_warn("Unable to open \"%s\": %s\n", config->pgp_certificate,
               g_strerror(errno));
        return -1;
    }

    gpg_type.data = g_malloc(sb.st_size);
    gpg_type.len = sb.st_size;

    if (read(fd, gpg_type.data, gpg_type.len) != gpg_type.len) {
        l_warn("read(%s) failed: %s\n", config->pgp_certificate,
               g_strerror(errno));
        close(fd);
        return -1;
    }
    close(fd);

    auth_register(&gpg_type);
    return 0;
}
