/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <openssl/blowfish.h>

#define LOG_FLAGS 0
#include "log.h"
#include "plugin.h"
#include "cipher.h"
#include "wire.h"
#include "debug.h"


/*
 * this implements the simple army/navy three rotor enigma engine.  
 * while extending it to the four rotor navy enigma it could be 
 * properly extended to support arbitrary combinations of rotors
 * who may rotate or not, but why bother?
 *
 * there's also ridiculous optimizations to be made:  
 * 	packing lookup tables down to their relevant bit widths
 *	simd instructions to do parallel lookups, absurd.
 *
 * the only test vectors I have are 
 *  rotors core ring  plain      cipher
 *  123    AAA  AAA   ABCDEFG    BJELRQZ
 *  123    KDT  TLP   HELLOWORLD OLYODHKPMA
 */

#define REFLECTOR 0
#define LEFT_ROTOR 1
#define RIGHT_ROTOR 3

#define NR_ROTORS (RIGHT_ROTOR + 1)

/* 
 * the while is 50% faster than %.. 
 */
#define CLAMP(foo, bar) do { while (foo > (bar-1) ) foo -= bar; } while (0)

struct rotor {
    unsigned char *mapping, notch;
	unsigned char forward[26], reverse[26];
} rotors[] = {
	{"EKMFLGDQVZNTOWYHXUSPAIBRCJ", 'Q', }, 
	{"AJDKSIRUXBLHWTMCQGZNPYFVOE", 'E', }, 
	{"BDFHJLCPRTXVZNYEIWGAKMUSQO", 'V', }, 
	{"ESOVPZJAYQUIRHXLNFTGKDCMWB", 'J', }, 
	{"VZBRGITYUPSDNHLXAWMJQOFECK", 'Z', }, 
	/* reflectors are dummy rotors witno rotating */
	{"YRUHQSLDPXNGOKMIEBFZCWVJAT", '.', },
	{"FVPJIAOYEDRZXWGCTKUQSBNMHL", '.', },
    {NULL,}
};

struct enigma_state {
    struct rotor_state {
        struct rotor *rot;
        unsigned char ring, core, starting_core;
    } rs[NR_ROTORS];
};

/*
 * have a signal pass through a rotor at the 'c' position.  core and
 * ring will be 0 for reflectors.. 
 */
char enigma_route(unsigned char c, struct rotor_state *rs, unsigned char *dir)
{
    c = c + rs->core - rs->ring + 26;
    CLAMP(c, 26);
    c = dir[c];
    c = c + rs->ring - rs->core + 26;
    CLAMP(c, 26);

	return c;
}

/*
 * the 'double roll-over' of interior rotors makes sense once you realize that
 * a rotor flips when one of the T bars is aligned with its notch _or_ its
 * righthand neighbour's notch.
 */
static void 
enigma_advance(struct enigma_state *es)
{
	int advance[NR_ROTORS] = {0,};
	int i;

	advance[RIGHT_ROTOR] = 1;

	for ( i = RIGHT_ROTOR; i > LEFT_ROTOR ; i-- ) {

		if ( es->rs[i].rot->notch == es->rs[i].core + 'A') {
			advance[i] |= 1;
			advance[i - 1] |= 1;
		}
	}
	for ( i = LEFT_ROTOR; i <= RIGHT_ROTOR ; i++ ) {

		if ( advance[i] ) {
			es->rs[i].core++;
            CLAMP(es->rs[i].core, 26);
		}
	}
}

static unsigned char 
enigma_keypress(unsigned char c, struct enigma_state *es)
{
    int i;
    enigma_advance(es);

    for ( i = RIGHT_ROTOR; i >= REFLECTOR ; i-- )
        c = enigma_route(c, &es->rs[i], es->rs[i].rot->forward);

    for ( i = LEFT_ROTOR; i <= RIGHT_ROTOR ; i++ )
        c = enigma_route(c, &es->rs[i], es->rs[i].rot->reverse);

    return c;
}

static void 
enigma_translate(struct cipher *ciph,
                    char *in, int in_len,
                    char *out, int out_len)
{
    struct enigma_state *es = ciph->key;
    int i;
    
    for ( i = LEFT_ROTOR ; i <= RIGHT_ROTOR ; i++)
        es->rs[i].core = es->rs[i].starting_core;

    /*
     * stupid enigma and its stupid 26 inputs and outputs.. we use
     * the stream of nibbles as letters, which means we're only
     * using A -> M.  tee hee :)
     */
    for ( i = 0 ; i < in_len ; i++) {
        out[i] = enigma_keypress((unsigned char)in[i] & 0xf, es);
        out[i] |= enigma_keypress((unsigned char)in[i] >> 4, es) << 4;
    }
}

static void 
enigma_schedule(struct cipher *ciph, char *data, int len)
{
    struct enigma_state *es = ciph->key;
    int offset = 0; /* cheesy, but whatever :) */
    int i;

    es->rs[REFLECTOR].rot = &rotors[5 + (data[offset++ % len] & 1)];

    for ( i = LEFT_ROTOR ; i <= RIGHT_ROTOR ; i++ ) {

        es->rs[i].rot = &rotors[data[offset++ % len] % 5];

        es->rs[i].starting_core = (data[offset++ % len] % 26);
        es->rs[i].ring = (data[offset++ % len] % 26);
    }
}

static int 
enigma_init(struct cipher *ciph)
{
    struct enigma_state *es;

    if ( (es = g_new0(struct enigma_state, 1)) == NULL)
        return -1;

    ciph->key = es;

    return 0;
}

static void
enigma_free(struct cipher *ciph)
{
    struct enigma_state *es = ciph->key;

    free(es);
}


struct cipher_funcs enigma_cf = {
    desc:       "3 rotor Enigma",
    encrypt:    enigma_translate,
    decrypt:    enigma_translate,
    schedule:   enigma_schedule,
    init:       enigma_init,
    free:       enigma_free,
};

int init_plugin(struct plugin_data *data, struct config *config)
{
    int i;
    unsigned char c;

    if ( cipher_register(CIPHER_ENIGMA, &enigma_cf) )
        return -1;

	for ( i = 0; rotors[i].mapping ; i++) {
		struct rotor *r = &rotors[i];

		for(c = 0; c < 26; c++) {

			r->forward[c] = r->mapping[c] - 'A';
			r->reverse[r->forward[c]] = c;
        }
    }

    return 0;
}
