/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <openssl/evp.h>

#define LOG_FLAGS 0
#include "log.h"
#include "plugin.h"
#include "digest.h"
#include "wire.h"
#include "debug.h"
#include "util.h"

static void 
evp_update(struct digest *dig, void *data, int len)
{
    EVP_MD_CTX *ctx = dig->state;

    EVP_DigestUpdate(ctx, data, len);
}

static int
evp_complete(struct digest *dig, void *out)
{
    int s;
    EVP_MD_CTX *ctx = dig->state;

    EVP_DigestFinal(ctx, out, &s);
    EVP_DigestInit(ctx, EVP_MD_CTX_md(ctx));

    return s;
}

static int 
evp_init(struct digest *dig)
{
    EVP_MD_CTX *ctx;
    EVP_MD *evp_type;

    switch(dig->type) {
        case DIGEST_MD5:
            evp_type = EVP_md5();
            break;
        case DIGEST_SHA1:
            evp_type = EVP_sha1();
            break;
        default:
            l_warn("got a digest type %d we didn't register?\n", dig->type);
            return - 1;
            break;
    }

    if ( (ctx = g_new0(EVP_MD_CTX, 1)) == NULL)
        return -1;

    EVP_DigestInit(ctx, evp_type);

    dig->max_output = EVP_MD_size(evp_type);
    dig->state = ctx;

    return 0;
}

static void
evp_free(struct digest *dig)
{
    EVP_MD_CTX *ctx = dig->state;

    /* 
     * are we just supposed to free() this?  who the hell knows?
     * yay for openssl!
     */
    if ( ctx ) {
        free(ctx);
        dig->state = NULL;
    }
}

struct digest_funcs evp_df = {
    desc:       "OpenSSL EVP digest wrapper",
    update:     evp_update,
    complete:   evp_complete,
    init:       evp_init,
    free:       evp_free,
};

int init_plugin(struct plugin_data *data, struct config *config)
{
    int types[] = {
        DIGEST_SHA1,
        DIGEST_MD5
    };
    int i;
    int count = 0;
        
    /* 
     * EVP_DigestFinal can't write a partial md, and we don't want to
     * be memcpying around, so holler if callers won't be able to
     * deal with our maximal digest len..
     */
    if ( MAX_DIGEST_BYTES < EVP_MAX_MD_SIZE ) {
        l_error("MAX_DIGEST_BYTES (%d) < EVP_MAX_MD_SIZE (%d)\n",
                MAX_DIGEST_BYTES, EVP_MAX_MD_SIZE);
        return -1;
    }

    for ( i = 0; i < ARRAY_SIZE(types) ; i++) {
        if ( digest_register(types[i], &evp_df) ) {
            l_debug("couldn't register digest %d!\n", types[i]);
            continue;
        }
        count++;
    }

    if ( count ) {
        OpenSSL_add_all_digests();
        return 0;
    }

    return -1;
}
