/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>

#include <openssl/dh.h>
#include <openssl/bn.h>

#include "log.h"
#include "wire.h"
#include "pkt.h"
#include "secret.h"
#include "debug.h"

/*
 * code for negotiating shared secrets that are kept from people that
 * might be passively monitoring the negotiation..
 *
 * secret agreement is cpu-hungry, we make it async so that other cpus or hw 
 * can do the work for us and let us know when its done..
 *
 * for now we can only do diffie-hellman, that should probably be exposed
 * via a module api..
 */

/*
 * 2048bit modulus from draft-ietf-ipsec-skip-07.txt
 */
static char p_hex[]=
    "F64257B7087F081772A2BAD6A942F305E8F95311394FB6F16EB94B3820DA01A7"
    "56A314E98F4055F3D007C6CB43A994ADF74C648649F80C83BD65E917D4A1D350"
    "F8F5595FDC76524F3D3D8DDBCE99E1579259CDFDB8AE744FC5FC76BC83C54730"
    "61CE7CC966FF15F9BBFD915EC701AAD35B9E8DA0A5723AD41AF0BF4600582BE5"
    "F488FD584E49DBCD20B49DE49107366B336C380D451D0F7C88B31C7C5B2D8EF6"
    "F3C923C043F0A55B188D8EBB558CB85D38D334FD7C175743A31D186CDE33212C"
    "B52AFF3CE1B1294018118D7C84A70A72D686C40319C807297ACA950CD9969FAB"
    "D00A509B0246D3083D66A45D419F9C7CBD894B221926BAABA25EC355E9320B3B" ;
static char g_hex[]="2";

static DH *dh;

/*
 * append the list of secrets we understand
 */
int
secret_append_list(struct packet *pkt)
{
    guint16 type;

    type = htons(B_SEC_DH);

    memcpy( pkt_push_tail(pkt, sizeof(type)),
            &type, sizeof(type));

    return 0;
}
/*
 * chose a secret negotiation method that we prefer..
 */
int
secret_choose(guint16 *list, int num)
{
    while(num--) {
        guint16 local;
        memcpy(&local, &list[num], sizeof(guint16));
        /*
         * only have DH for now, la la la
         */
        if ( ntohs(local) == B_SEC_DH ) 
            return B_SEC_DH;
    }

    return -1;
}

struct secret *
secret_alloc(int len)
{
    struct secret *sec;

    /* 
     * important that this is 0ed so that the hash gets intialized
     * to 0.
     */
    if ( (sec = calloc(sizeof(struct secret) + len, 1)) == NULL) {
        l_warn("couldn't allocate secret\n");
        return NULL;
    }

    sec->len = len;

    return sec;
}

void
secret_free(struct secret *sec)
{
    free(sec);
}

/*
 * append our public side of the secret material to the end
 * of the packet.
 */
int
secret_append_ann(struct packet *pkt, guint16 type)
{
    int len;
    char *tail;

    if ( type != B_SEC_DH )
        return -1;

    len = BN_num_bytes(dh->pub_key);
    l_debug("appending pub key of %d bytes\n", len);
    tail = pkt_push_tail(pkt, len);
    BN_bn2bin(dh->pub_key, tail);

    return 0;
}

/*
 * take a public key read off the wire and combine
 * it with our private stuff to get the shared secret 
 */
struct secret *
secret_parse_ann(struct packet *pkt, guint16 type)
{
    int len;
    BIGNUM *bn;
    struct secret *sec;

    if ( type != B_SEC_DH ) {
        l_warn("type %d != B_SEC_DH\n", type);
        return NULL;
    }

    len = DH_size(dh);

    if ( pkt->len < len ) {
        l_warn("pkt->len %d < dh public key len %d\n", pkt->len, len);
        return NULL;
    }

    if ( (bn = BN_bin2bn(pkt->head, len, NULL)) == NULL) {
        l_warn("bin2bin failed?\n");
        return NULL;
    }

    if ( (sec = secret_alloc(len)) == NULL) {
        l_warn("couldn't alloc secret of len %d\n", len);
        goto out;
    }
         
    DH_compute_key(sec->data, bn, dh);

    hexdump(sec->data, len, "found secret\n");

out:
    BN_free(bn);
    return sec;
}

int
secret_init(void)
{
    /*
     * generate our private/public key pair.
     */
    dh = DH_new();
    BN_hex2bn(&dh->p, p_hex);
    BN_hex2bn(&dh->g, g_hex);

    DH_generate_key(dh);

    return 1;
}
