/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2001 Phil Schwan <phil@off.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#include <glib.h>

#include "log.h"
#include "shell.h"

/*
 * Provide infrastructure for shell-script-calling auth types.
 */

struct shell_handle *
shell_start(const char *file, char *const argv[])
{
    struct shell_handle *handle;

    handle = g_new0(struct shell_handle, 1);
    if (!handle)
        return NULL;

    signal(SIGPIPE, SIG_IGN);

    pipe(handle->stdin);
    pipe(handle->stdout);
    pipe(handle->stderr);
    handle->pid = fork();
    if (handle->pid == 0) {
        /* Child */
        close(handle->stdin[1]);
        close(handle->stdout[0]);

        if (dup2(handle->stdin[0], 0) != 0)
            l_critical("stdin dup2 failed: %d (%s)\n", 
                    errno, g_strerror(errno));

        else if (dup2(handle->stdout[1], 1) != 1)
            l_critical("stdout dup2 failed: %d (%s)\n", 
                    errno, g_strerror(errno));

        else if (dup2(handle->stderr[1], 2) != 2)
            l_critical("stderr dup2 failed: %d (%s)\n", 
                    errno, g_strerror(errno));

        else if (execvp(file, argv) != 0)
            l_critical("execvp(%s) failed: %d (%s)\n",
                    file, errno, g_strerror(errno));

        exit(1);
    } else {
        close(handle->stdin[0]);
        close(handle->stdout[1]);
        close(handle->stderr[1]);
    }

    return handle;
}

int
shell_write(struct shell_handle *handle, const void *buf, size_t count)
{
    int retval;

    if (count == 0) {
        /* A 'count' of zero means "I'm finished writing, close the FD" */
        close(handle->stdin[1]);
        handle->stdin[1] = -1;
        return 0;
    }

    retval = write(handle->stdin[1], buf, count);
    if (retval < 0) {
        l_warn("write of %d bytes to fd %d failed: %d (%s)\n",
               count, handle->stdin[1], errno, g_strerror(errno));
    } else if (retval < count) {
        l_warn("short write (%d of %d bytes)\n", retval, count);
    }

    return retval;
}

int
shell_read_stdout(struct shell_handle *handle, void *buf, size_t count)
{
    int retval;

    if (count == 0) {
        /* A 'count' of zero means "I'm finished reading, close the FD".
         * FIXME: should this optionally kill the process?  Is that useful? */
        close(handle->stdout[0]);
        handle->stdout[0] = -1;
        return 0;
    }

    retval = read(handle->stdout[0], buf, count);
    if (retval < 0) {
        l_warn("read of %d bytes from fd %d failed: %d (%s)\n",
               count, handle->stdout[0], errno, g_strerror(errno));
    }
    return retval;
}

int
shell_read_stderr(struct shell_handle *handle, void *buf, size_t count)
{
    int retval;

    if (count == 0) {
        /* Blah blah blah, same old story */
        close(handle->stderr[0]);
        handle->stderr[0] = -1;
        return 0;
    }

    retval = read(handle->stderr[0], buf, count);
    if (retval < 0) {
        l_warn("read of %d bytes from fd %d failed: %d (%s)\n",
               count, handle->stderr[0], errno, g_strerror(errno));
    }
    return retval;
}

void
shell_kill(struct shell_handle *handle)
{
    kill(handle->pid, SIGTERM);
}

/* Returns the return code of the process if it exited normally, or -1 if it
 * exited abnormally (from a signal, say).  I know that -1 overlaps.  If that
 * causes trouble, I'll change it. */
int
shell_finish(struct shell_handle *handle)
{
    int status;

    if (handle->stdin[1] != -1)
        close(handle->stdin[1]);
    if (handle->stdout[0] != -1)
        close(handle->stdout[0]);
    if (handle->stderr[0] != -1)
        close(handle->stderr[0]);

    if (handle->pid > 0)
        waitpid(handle->pid, &status, 0);

    g_free(handle);

    if (WIFEXITED(status))
        return WEXITSTATUS(status);
    else
        return -1;
}
