/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * this file implements the ospf-like topology flooding.  It minimally
 * effects the other code, it just needs to know when links go up 
 * or down and needs a way to send packets to the other side of them.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include <glib.h>

#include "config.h"
#include "log.h"
#include "util.h"
#include "udp.h"
#include "link.h"
#include "wire.h"
#include "debug.h"
#include "outgoing.h"

/*
 * an instance of an lsa in memory. 
 */
struct lsa {
    struct topo_state *ts; /* how irritating.. expiry needs to know its owner */
    struct sockaddr_in orig;
    struct sockaddr_in sender;
    guint32 seq;
    struct packet *pkt;
    gint expiry_id;
    refcount_t __refcount;
    void (*__destroy)(struct lsa *lsa);
};

/*
 * the state we hold while we're waiting for an ack of a 
 * lsa we sent to a neighbour.   there is one of these for originating
 * node's lsa that we send down a link..
 */
struct pending_ack {
    struct sockaddr_in orig;
    struct link *link;
    guint32  sent;   /* XXX should be bignum stuff.. */
    guint timer_id;
};

static void
lsa_destroy(struct lsa *lsa)
{
    l_debug("destroying lsa %p\n", lsa);

    if (lsa->pkt)
        ref_release(lsa->pkt);

    free(lsa);
}


static struct lsa *
lsa_getref_alloc(void)
{
    struct lsa *lsa = g_new0(struct lsa, 1);

    if (lsa == NULL)
        return NULL;

    l_debug("allocating lsa %p\n", lsa);

    REF_INIT(lsa, lsa_destroy);

    return lsa;
}

/*
 * callers are responsible for getting the pa out of the hash, 
 * and should set timer_id to 0 if they remove the timer..
 */
static void 
pa_destroy(gpointer data)
{
    struct pending_ack *pa = data;

    pa->link = NULL;

    if(pa->timer_id) { 
        g_source_remove(pa->timer_id);
        pa->timer_id = 0;
    }

    l_debug("freeing pa %s:%d[%d]\n",inet_ntoa(pa->orig.sin_addr),
            ntohs(pa->orig.sin_port), pa->sent);

    free(pa);
}

/*
 * this is called to originally send an lsa and also when we don't
 * get an ack before the timer expires.. we try to ack forever, if
 * the link is really hosed the keep-alives will find that out and
 * tear it down, calling topo_link_down..
 *
 * this is the only async code path in the topology goo.  to make
 * our lives easier we store a reference to the lsa via the 
 * originating address and look it up in the lsa tree.. we'll send
 * out a new lsa if it is updated under us and will silently
 * stop if the lsa is axed..
 *
 * we won't get here if the link is removed, the timer will
 * have been shutoff via the call to topo_link_down()
 */
static gboolean
lsa_retransmit(gpointer data)
{
    struct pending_ack *pa = data;
    struct link *link = pa->link;
    struct socket *sk = link->sk;
    struct lsa *lsa;

    if ((lsa = g_tree_lookup(sk->ts.lsa_tree, &pa->orig)) == NULL) {
        l_debug("faild to retransmit lsa %s:%d[%d]\n",
                inet_ntoa(pa->orig.sin_addr), ntohs(pa->orig.sin_port), 
                pa->sent);
        pa->timer_id = 0;
        g_hash_table_remove(link->lt.pending_acks, &pa->orig);
        return FALSE;
    }

    pa->sent = lsa->seq;

    l_debug("sending lsa %s:%d[%d]\n",inet_ntoa(pa->orig.sin_addr),
            ntohs(pa->orig.sin_port), pa->sent);

    outgoing_link(lsa->pkt, link, LINK_LSA, 0);

    return TRUE;
}

/*
 * this is called by both code paths that want to send individual
 * lsa's over a link.  
 */
static void 
initiate_send(struct link *link, struct lsa *lsa)
{
    struct pending_ack *pa;
    struct link_topo *lt = &link->lt;

    /*
     * It would be dreadfully embarrassing if we forwarded an lsa
     * to someone who already gave it to us or who generated it
     * in the first place.
     *
     * what would the neighbours think?
     */
    if (    sin_equal(&lsa->orig, &link->sin)  ||
            sin_equal(&lsa->sender, &link->sin)      )
        return;

    /*
     * look to see if we've sent out any lsa for the originator
     * of this lsa over this link..
     */
    pa = g_hash_table_lookup(lt->pending_acks, &lsa->orig);

    if (pa) { /* rtlider */
        if (pa->sent > lsa->seq) {
            l_debug("very strange, pa->sent > lsa->seq\n");
            return;
        }

        if (pa->sent == lsa->seq) {
            l_debug("already waiting for ack to %s:%d[%d]\n",
                    inet_ntoa(pa->orig.sin_addr), ntohs(pa->orig.sin_port), 
                    pa->sent);
            return;
        }
        /*
         * fall out for case where we're waiting for
         * an ack for an out of date lsa.. we want to send the new 
         * one now..
         */
    } else {
        if ((pa = g_new0(struct pending_ack, 1)) == NULL)
            return;

        l_debug("allocated a new pa %p\n", pa);

        memcpy(&pa->orig, &lsa->orig, sizeof(struct sockaddr_in));
        pa->link = link;
        g_hash_table_insert(lt->pending_acks, &pa->orig, pa);
    }

    /*
     * make sure the retransmit timer doesn't fire too early..
     */
    if (pa->timer_id) {
        g_source_remove(pa->timer_id);
    }
    pa->timer_id = g_timeout_add(lsa->ts->lsa_retrans_timeout,
                                 lsa_retransmit, pa);

    lsa_retransmit(pa); /* kick the packet now */
}

static void
got_an_ack(struct w_lsa_header *lh, struct link *link)
{
    struct pending_ack *pa;
    struct sockaddr_in sin;

    sin_init(&sin);
    sin.sin_addr.s_addr = lh->orig_ip;
    sin.sin_port = lh->orig_port;

    if ((pa = g_hash_table_lookup(link->lt.pending_acks, &sin)) == NULL) {
        l_debug("unexpected ack %s:%d[%d]\n",inet_ntoa(link->sin.sin_addr),
                ntohs(link->sin.sin_port), ntohl(lh->seq));
        return;
    }

    if (ntohl(lh->seq) > pa->sent) {
        l_debug("ack %s:%d[%d] is greater than pa->sent %d ?\n",
                inet_ntoa(pa->orig.sin_addr), ntohs(pa->orig.sin_port), 
                ntohl(lh->seq), pa->sent);
        return;
    }

    if (ntohl(lh->seq) < pa->sent) {
        /* not good enough.. */
        l_debug("got a short ack %s:%d[%d]\n", inet_ntoa(pa->orig.sin_addr), 
                ntohs(pa->orig.sin_port), ntohl(lh->seq));
        return;
    }

    l_debug("ok ack %s:%d[%d]\n", inet_ntoa(pa->orig.sin_addr), 
            ntohs(pa->orig.sin_port), pa->sent);

    /*
     * yay, we can shut up now
     */
    g_hash_table_remove(link->lt.pending_acks, &pa->orig);
}

static void
send_link_lsa(gpointer key, gpointer value, gpointer user_data)
{
    struct link *link = value;
    struct lsa *lsa = user_data;

    initiate_send(link, lsa);
}

static gboolean
send_lsa_link(gpointer key, gpointer value, gpointer user_data)
{
    struct link *link = user_data;
    struct lsa *lsa = value;

    initiate_send(link, lsa);

    return FALSE;
}

static gboolean 
file_write_lsa(gpointer key, gpointer value, gpointer data)
{
    struct lsa *lsa = value;
    int fd = (int)data;

    write(fd, lsa->pkt->head, lsa->pkt->len);

    return FALSE;
}

/*
 * update the file that stores all the lsas..
 */
static gboolean
update_lsa_file(gpointer data)
{
    struct topo_state *ts = data;
    char tmp[PATH_MAX], *filename = ts->lsa_file;
    int fd;

    ts->file_timeout_id = 0;

    strncpy(tmp, filename, PATH_MAX - 5);
    strcat(tmp, ".tmp");

    if ((fd = open(tmp, O_CREAT|O_RDWR), 0700) < 0) {
        l_warn("couldn't open lsa tmp file %s: %s\n",
                filename, strerror(errno));
        return FALSE;
    }

    g_tree_foreach(ts->lsa_tree, file_write_lsa, (gpointer)fd);

    close(fd);
    unlink(filename);
    rename(tmp, filename);
    
    return FALSE;
}

static int
init_lsa_file(struct topo_state *ts, char *filename)
{
    int fd;

    if ( filename == NULL) {
        l_critical("no lsa file specified in the config file\n");
        return 0;
    }

    if ((fd = open(filename, O_CREAT|O_RDWR, 0600)) < 0) {
        l_warn("couldn't open lsa file %s: %s\n",
                filename, strerror(errno));
        return 0;
    }

    close(fd);
    ts->lsa_file = filename;

    return 1;
}


/*
 * remove an lsa from the tree.  called to forcefully remove stale lsas,
 * but also from an lsa's expiry timer.  since its called from the timer,
 * it can't remove the timer when its called from try_tree_update..
 */
static gboolean
lsa_remove(gpointer data)
{
    struct lsa *lsa = data;

    l_debug("trying to remove lsa %p\n", lsa);

    g_tree_remove(lsa->ts->lsa_tree, &lsa->orig);
    lsa->expiry_id = 0;
    ref_release(lsa);

    return FALSE;
}
/*
 * someone's given us an authentic lsa.. if its a new lsa we need
 * to put it in the tree and forward it on to our neighbours who
 * might not have seen it yet.
 */
static void
try_tree_update(struct socket *sk, struct topo_state *ts, struct lsa *lsa)
{
    struct lsa *old_lsa;

    lsa->ts = ts;
    old_lsa = g_tree_lookup(ts->lsa_tree, &lsa->orig);

    if (old_lsa) {
        /*
         * we already have a more up to date lsa for this
         * origintator
         */
        if (old_lsa->seq >= lsa->seq) {
            return;
        }

        /*
         * destroy the out of date lsa
         */
        g_source_remove(old_lsa->expiry_id);
        lsa_remove(old_lsa);
    }

    g_tree_insert(ts->lsa_tree, &lsa->orig, ref_get(lsa));
    lsa->expiry_id = g_timeout_add(ts->lsa_expiry_timeout, lsa_remove, lsa);

    /*
     * delay file syncing just a bit so we can batch lots of updates..
     */
    if ( ts->file_timeout_id == 0)
        ts->file_timeout_id = g_timeout_add(2000, update_lsa_file, ts);

    /*
     * all right, we got a new lsa, lets spray it to all our peers
     */
    g_hash_table_foreach(sk->link_nei_hash, send_link_lsa, lsa);
}


static void
append_nei(gpointer key, gpointer value, gpointer user_data)
{
    struct link *link = value;
    struct w_lsa_nei **wln = user_data;

    (*wln)->nei_ip = link->sin.sin_addr.s_addr;
    (*wln)->nei_port = link->sin.sin_port;

    (*wln)++;
}
/*
 * this updates our local lsa and tries to shove it into the 
 * tree of all the lsas we know about in the network.  Its called when
 * our link count changes or via the refresh timer...
 */
static gboolean 
build_our_lsa(gpointer data)
{
    struct socket *sk = data;
    struct packet *pkt;
    struct w_lsa_header *lh;
    struct w_lsa_payload lp;
    struct w_lsa_nei *wln;
    struct lsa *lsa;
    struct topo_state *ts = &sk->ts;
    int num_neigh, nei_bytes;

    if ((lsa = lsa_getref_alloc()) == NULL)
        return TRUE;

    num_neigh = g_hash_table_size(sk->link_nei_hash);
    nei_bytes = num_neigh * sizeof(struct w_lsa_nei);
    lp.num_neighbours = htons(num_neigh);
    lp.data_len = htons(0);

    if ((pkt = pkt_getref_alloc(sizeof(struct w_lsa_header),
                    sizeof(struct w_lsa_payload) + nei_bytes)) == NULL)
        goto out;

    ts->lh.seq = htonl(ntohl(ts->lh.seq) + 1);

    lh = pkt_pull_head(pkt, sizeof(*lh));
    memcpy(lh, &ts->lh, sizeof(*lh));

    memcpy( pkt_push_tail(pkt, sizeof(lp)), 
            &lp, sizeof(lp));

    wln = pkt->tail;
    pkt_push_tail(pkt, nei_bytes);
    g_hash_table_foreach(sk->link_nei_hash, append_nei, &wln);

    lsa->pkt = pkt;
    lsa->sender.sin_addr.s_addr = INADDR_NONE;
    sin_init(&lsa->orig);
    lsa->orig.sin_addr.s_addr = ts->lh.orig_ip;
    lsa->orig.sin_port = ts->lh.orig_port;
    lsa->seq = ntohl(ts->lh.seq);

    l_debug("built lsa %d\n", lsa->seq);

    try_tree_update(sk, ts, lsa);

out:
    ref_release(lsa);
    return TRUE;
}

/*
 * see if our incoming packet is new and interesting.  notice that
 * we steal the packet from the incoming stack.. 
 */
static void
got_an_lsa(struct packet *pkt, struct link *link)
{
    struct w_lsa_header lh;
    struct lsa *lsa;
    int slack;

    /*
     * quickly send out the ack, outgoing_link() must be non-destructive
     */
    slack = pkt->len - sizeof(lh);
    pkt_pull_tail(pkt, slack);
    outgoing_link(pkt, link, LINK_LSA_ACK, 0);
    pkt_push_tail(pkt, slack);

    if ((lsa = lsa_getref_alloc()) == NULL)
        return;

    memcpy(&lh, pkt->head, sizeof(lh));

    lsa->pkt = ref_get(pkt);
    lsa->sender.sin_addr.s_addr = link->sin.sin_addr.s_addr;
    lsa->sender.sin_port = link->sin.sin_port;
    sin_init(&lsa->orig);
    lsa->orig.sin_addr.s_addr = lh.orig_ip;
    lsa->orig.sin_port = lh.orig_port;
    lsa->seq = ntohl(lh.seq);

    try_tree_update(link->sk, &link->sk->ts, lsa);

    ref_release(lsa);
}

void
topo_incoming(struct packet *pkt, struct link *link, int type)
{
    struct w_lsa_header lh;

    if (pkt->len < sizeof(lh)) {
        l_debug("pkt len %d not big enough for %d byte w_lsa_header\n",
                pkt->len, sizeof(lh));
        return;
    }

    hexdump(pkt->head, pkt->len, "got topo packet\n");

    memcpy(&lh, pkt->head, sizeof(lh));

    switch (type) {
        case LINK_LSA: 
            got_an_lsa(pkt, link);
            break;

        case LINK_LSA_ACK:
            /*
             * we trust the link layer to have verified the integrity of this..
             */
            got_an_ack(&lh, link);
            break;
    }
}

/*
 * called for all new links..
 */
int 
topo_link_up(struct socket *sk, struct link *link) 
{
    if (! link->neighbour)
        return 0;

    if (link->lt.pending_acks == NULL) {
        link->lt.pending_acks = g_hash_table_new_full(sin_hash, sin_equal,
                                                NULL, pa_destroy);
        if(link->lt.pending_acks == NULL) {
            l_debug("couldn't init link's topo pending hash\n");
            return -1;
        }
    }

    /*
     * update our lsa and send it to our neighbours..
     */
    build_our_lsa(sk);

    /*
     * also send all our lsas to the new neighbour so he can get
     * started..
     */
    g_tree_foreach(sk->ts.lsa_tree, send_lsa_link, link);

    return 0;
}

/*
 * called when a link is being torn down.  we'll have some cleaning
 * up to do..
 */
void
topo_link_down(struct link *link) 
{
    if (link->lt.pending_acks == NULL) 
        return;

    /*
     * calls our value teardown func..
     */
    g_hash_table_destroy(link->lt.pending_acks);
    link->lt.pending_acks = NULL;
}

/*
 * called on socket buildup.. we don't have a matching destroy
 * to be called when a socket is destroyed yet..  XXX so this will
 * leak stuff if a socket buildup error happens after the socket buildup
 * has called topo_init :/
 */
int
topo_init(struct daemon *dae, struct socket *sk)
{
    struct topo_state *ts = &sk->ts;
    struct w_lsa_header *lh = &ts->lh;

    if ((ts->lsa_tree = g_tree_new(sin_compare)) == NULL) {
        l_critical("couldn't initialize topo lsa tree\n");
        return -1;
    }

    if ( ! init_lsa_file(ts, dae->config->lsa_file)) { 
        g_tree_destroy(ts->lsa_tree);
        return -1;
    }

    lh->orig_ip = sk->sin.sin_addr.s_addr;
    lh->orig_port = sk->sin.sin_port;
    lh->seq = htonl(0);

    ts->lsa_retrans_timeout = dae->config->lsa_retrans_timeout * 1000;
    ts->lsa_expiry_timeout = dae->config->lsa_expiry_timeout * 1000;
    ts->lsa_refresh_id = g_timeout_add(dae->config->lsa_refresh_interval * 1000,
                                       build_our_lsa, sk);

    return 0;
}

struct topo_walker {
    void *buf;
    int remaining;
};
static gboolean 
add_lsa_len(gpointer key, gpointer value, gpointer data)
{
    struct lsa *lsa = value;
    struct topo_walker *tw = data;

    if ( tw->remaining >= lsa->pkt->len ) {
        memcpy(tw->buf, lsa->pkt->head, lsa->pkt->len);
        tw->buf += lsa->pkt->len;
        tw->remaining -= lsa->pkt->len;
    } else 
        tw->remaining = -1;

    return FALSE;
}
int
topo_show(struct daemon *dae, void *buf, size_t len)
{
    GList *list;
    struct topo_walker tw;

    tw.remaining = len;
    tw.buf = buf;

    for ( list = dae->socket_list ; list ; list = g_list_next(list) ) {
        struct socket *sk = list->data;
        struct topo_state *ts = &sk->ts;

        g_tree_foreach(ts->lsa_tree, add_lsa_len, &tw);
    }

    if ( tw.remaining < 0 )
        return -ENOSPC;

    return len - tw.remaining;
}
