/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <net/if.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <linux/if_tun.h>  /* ugh */

#include "log.h"
#include "tun.h"
#include "tunnel.h"
#include "incoming.h"

static int socket_fd = -1;

static void 
setup_socket(void)
{
    if ( socket_fd >= 0)
        return;

    socket_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);

    if ( socket_fd < 0 ) 
        l_debug("had trouble allocting dev conf socket: %s\n",
                strerror(errno));
}

/*
 * int_*() could be pulled into an abstracted api when we want to support
 * more than just /dev/net/tun .. for now its enough.
 */

void 
int_destroy(struct interface *intf)
{
    l_debug("destroying intf %s\n", intf->dev);
    /*
     * this mirrors the order that things are built up in
     */
    if ( intf->source ) {
        if ( intf->pfd.events ) 
            g_source_remove_poll(intf->source, &intf->pfd);
        g_source_destroy(intf->source);
        if ( intf->pfd.fd > 0 ) 
            close(intf->pfd.fd);
    }
    free(intf);
}

static gboolean 
source_prepare(GSource *source, gint *timeout)
{
    *timeout = -1; /* block forever */
    return FALSE;  /* we have to go to poll to find out if we're ready */
}

/* FIXME: GSource is technically marked private, but how else am I supposed to
 * do this?  They just mean ``don't change stuff directly,'' right? --phil 
 * I think it just might be the callback_ stuff at the top that is marked
 * private, but I'm probably wrong.  this seems dorky, none the less. --zach */
static gboolean 
source_check(GSource *source)
{
    GSList *l;

    for (l = source->poll_fds; l != NULL; l = l->next) {
        GPollFD *pfd = (GPollFD *)l->data;

        if (pfd->revents != 0)
            return 1;
    }

    return 0;
}

static gboolean 
source_dispatch(GSource *source, GSourceFunc callback, gpointer user_data)
{
    struct interface *intf = user_data;
    GPollFD *pfd = &intf->pfd;
    int toread = intf->mtu + 1;
    struct packet *pkt;
    ssize_t len;

    if ( (pkt = pkt_getref_alloc(128, toread)) == NULL) /* XXX 128? :) */
        return TRUE;

    len = read(pfd->fd, pkt->head, toread);

    if ( len <= 0 ) {
        l_debug("read error %d from dev %s: %s\n", len, 
                intf->dev, strerror(errno));
        goto out;
    }

    if ( len == toread ) {
        l_debug("read one byte over mtu %d on dev %s\n", intf->mtu, intf->dev);
        goto out;
    }

    pkt_push_tail(pkt, len);
    incoming_frame(intf->tunnel, pkt);

out:
    ref_release(pkt);

    return TRUE;
}

GSourceFuncs source_funcs = {
    prepare: source_prepare,
    check: source_check,
    dispatch: source_dispatch,
};

void 
int_deliver(struct interface *intf, struct packet *pkt)
{
    write(intf->pfd.fd, pkt->head, pkt->len);
}

/*
 * again, should be core code.
 */
static int 
int_setup(struct interface *intf, guint32 addr, guint32 netmask)
{
    struct ifreq ifr;
    struct sockaddr_in *sin;
    int flags;

    setup_socket();
    memset(&ifr, 0, sizeof(ifr));
    memcpy(ifr.ifr_name, intf->dev, IFNAMSIZ);

    /*
     * we need to take the interface down before reconfiguring
     * it
     */
    if ( ioctl(socket_fd, SIOCGIFFLAGS, &ifr) < 0) {
        l_debug("couldn't get flags from dev %s: %s\n",
                ifr.ifr_name, strerror(errno));
        return -1;
    }
    flags = ifr.ifr_flags & ~(IFF_UP|IFF_RUNNING);
    if ( ioctl(socket_fd, SIOCSIFFLAGS, &ifr) < 0) {
        l_debug("couldn't set flags for dev %s: %s\n",
                ifr.ifr_name, strerror(errno));
        return -1;
    }

    intf->mtu = 1300; /* XXX derive from underlying real device.. config? */
    ifr.ifr_mtu = intf->mtu;

    if ( ioctl(socket_fd, SIOCSIFMTU, &ifr) < 0)  {
        l_debug("couldn't set interface %s mtu to %d: %s\n",
                ifr.ifr_name, ifr.ifr_mtu, strerror(errno));
        return -1;
    }

    /*
     * ok, lets give it the address we negotiated for
     */
    sin = (struct sockaddr_in *)&ifr.ifr_addr;
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = addr;

    if ( ioctl(socket_fd, SIOCSIFADDR, &ifr) < 0)  { 
        l_debug("couldn't set int %s addr to %s: %s\n",
                intf->dev, inet_ntoa(sin->sin_addr),
                strerror(errno));
        return -1;
    }

    sin->sin_addr.s_addr = netmask;

    if ( ioctl(socket_fd, SIOCSIFNETMASK, &ifr) < 0)  { 
        l_debug("couldn't set int %s netmask to %s: %s\n",
                intf->dev, inet_ntoa(sin->sin_addr),
                strerror(errno));
        return -1;
    }

    /*
     * its ok to have some frames arrive before we add the source to glib..
     * it'll just have some pending the first time we hit the idle loop
     */
    ifr.ifr_flags = flags | (IFF_UP|IFF_RUNNING);
    if ( ioctl(socket_fd, SIOCSIFFLAGS, &ifr) < 0) {
        l_debug("couldn't bring up dev %s: %s\n", 
                ifr.ifr_name, strerror(errno));
        return -1;
    }

    return 0;
}

struct interface * 
int_allocate(struct tunnel *tunnel)
{
    struct interface *intf;
    struct ifreq ifr;

    memset(&ifr, 0, sizeof(ifr));

    if ((intf = g_new0(struct interface, 1)) == NULL) {
        l_debug("couldn't alloc interface\n");
        goto error;
    }

    if ( (intf->source = g_source_new(&source_funcs, sizeof(GSource))) == NULL) {
        l_debug("couldn't allocate source\n");
        goto error;
    }

    if ( (intf->pfd.fd = open("/dev/net/tun", O_RDWR)) < 0) {
        l_debug("couldn't open /dev/net/tun: %s\n", strerror(errno));
        goto error;
    }

    ifr.ifr_flags = IFF_TUN|IFF_NO_PI;

    if ( ioctl(intf->pfd.fd, TUNSETIFF, &ifr) < 0) {
        l_debug("couldn't set interface: %s\n", strerror(errno));
        goto error;
    }

    strncpy(intf->dev, ifr.ifr_name, IFNAMSIZ);
    intf->dev[IFNAMSIZ-1] = '\0';

    if ( int_setup(intf, tunnel->addr, tunnel->netmask) ) 
        goto error;

    intf->pfd.events = G_IO_IN | G_IO_HUP | G_IO_ERR;
    intf->tunnel = tunnel;

    g_source_add_poll(intf->source, &intf->pfd);
    g_source_set_priority(intf->source, G_PRIORITY_DEFAULT);
    g_source_set_callback(intf->source, NULL, intf, NULL);
    g_source_attach(intf->source, NULL);

    l_debug("opened tap %s\n", intf->dev);

    return intf;

error:
    if ( intf ) 
        int_destroy(intf);

    return NULL;
}
