/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * tunnel management.  negotiating tunnels, making sure they're still alive,
 * tearing them down, etc.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <glib.h>

#include "main.h"
#include "util.h"
#include "udp.h"
#include "topo.h"
#include "tunnel.h"
#include "pkt.h"
#include "wire.h"
#include "log.h"
#include "neg.h"
#include "seqwind.h"
#include "random.h"
#include "outgoing.h"
#include "defaults.h"
#include "link.h"
#include "debug.h"

#define TUNNEL_PARENT_SK(TUN) \
    ( (TUN)->link_reverse ? (TUN)->link_reverse->sk : (TUN)->link_forward->sk )

/*
 * boy, are tunnels ever irritating.
 *
 * tunnels are not symmetric, they're heavily directionally oriented.  the
 * client negotiates everything, all the server hops that make up the tunnel
 * are passive in the negotiation.  the client shares secrets with all
 * the hops in the tunnel, but the hops don't share tunnel secrets with
 * each other.. 
 *
 * a tunnel is built up iteratively.   the client bulds up the first
 * hop of the tunnel.. then, through this tunnel, it negotiates the 
 * second hop, etc.  only when the final hop is negotiated will
 * any of the hops let real traffic through.. until that point they'd
 * only allow negotiation traffic to be forwarded through the tunnel.
 *
 * our 'struct tunnel' really describes a node's hop in a tunnel.
 * we arbitrarily call traffic originating from the client 'forward'
 * and traffic heading to the client from the tunnel exit 'reverse'.
 *
 * client -> forward traffic -> hop -> forward traffic -> exit 
 * client <- reverse traffic <- hop <- reverse traffic <- exit 
 *
 * between any two hops we have two tunnel ids.  in hop negotiation
 * we send our local id to the remote guy, incoming packets from him
 * are then tagged with that id so we can lookup our state.  we send
 * things to the other side and tag it with their id..  these are
 * the local and remote ids, and they exist for the traffic that heads
 * towards our forward neighbour, and towards the previous.
 *
 * rekeying doesn't make sense for them right now.. that needs a way
 * to send a message to a specific hop.  this can be done by sending
 * a message that is propogated to all the hops until one of them
 * can verify it as being to them.. what a pain.
 */

/*
 * some quick simple routines for managing ranges of addresses.  this
 * probably wants to be a range tree, or something, should we really
 * let people negotiate networks.
 *
 * is anyone else noticing the similarity between this and the
 * seqwind bitmask code?
 */
static guint32 
ip_alloc(struct ip_block *ipb)
{
    int i, pos = 0;
    guint32 ret = 0;

    for ( i = 0; i < ipb->num_words ; i++) {
        
        if ( (pos = ffs(ipb->mask[i])) == 0 )
            continue;

        pos--;

        if ( pos >= ipb->num_addrs )  /* XXX pos * i*32, or something */
            break;

        ret = pos + ipb->base;
        ipb->mask[i] &= ~( 1 << pos );
        break;
    }

    l_debug("returning %x from pos %d base %x\n", ret, pos, ipb->base);

    return htonl(ret);
}

static void
ip_free(struct ip_block *ipb, guint32 addr)
{
    int offset, bit, *mask;

    addr = ntohl(addr);

    if ( addr < ipb->base || addr >= (ipb->base + ipb->num_addrs) ) {
        l_debug("trying to free addr %x out of range [%x - %x])\n", 
                htonl(addr), htonl(ipb->base), 
                htonl(ipb->base + ipb->num_addrs));
        return;
    }

    offset = addr - ipb->base;
    mask = ipb->mask +  ( offset / ( sizeof(ipb->mask[0]) * 8 ) );
    bit = 1 << ( offset % ( sizeof(ipb->mask[0]) * 8 ) );

    if ( *mask & bit ) 
        l_debug("trying to free free addr %x\n", htonl(addr));

    *mask |= bit;
}

static void
tunnel_destroy(struct tunnel *tunnel)
{
    struct list_head *pos, *n;

    l_debug("destroying tunnel %p\n", tunnel);

    if ( tunnel->addr ) {
        struct socket *sk = TUNNEL_PARENT_SK(tunnel);
        ip_free(sk->ipb, tunnel->addr);
        ip_free(sk->ipb, htonl(ntohl(tunnel->addr) + 1));
    }
    list_for_each_safe(pos,n, &tunnel->multi_keys_head) {
        struct keys *keys = list_entry(pos, struct keys, multi_entry);

        list_del(&keys->multi_entry);
        keys_destroy(keys);
    }

    if ( tunnel->session )
        secret_free(tunnel->session);

    if ( tunnel->intf )
        int_destroy(tunnel->intf);
    if ( tunnel->link_forward )
        ref_release(tunnel->link_forward);
    if ( tunnel->link_reverse )
        ref_release(tunnel->link_reverse);

    free(tunnel);
}

struct tunnel *
tunnel_getref_alloc(struct link *link_reverse, struct link *link_forward)
{
    struct tunnel *tunnel;

    if ( (tunnel=g_new0(struct tunnel, 1)) == NULL)
        return NULL;

    seq_init(&tunnel->end_seq);
    REF_INIT(tunnel, tunnel_destroy);

    INIT_LIST_HEAD(&tunnel->fwd_iterator);
    INIT_LIST_HEAD(&tunnel->rev_iterator);
    INIT_LIST_HEAD(&tunnel->multi_keys_head);

    if ( link_forward ) 
        tunnel->link_forward = ref_get(link_forward);

    if ( link_reverse )
        tunnel->link_reverse = ref_get(link_reverse);

    l_debug("allocated tunnel %p\n", tunnel);
    
    return tunnel;
}

struct tunnel *
tunnel_id_getref(struct link *link, guint32 id)
{
    struct tunnel *tunnel;

    if ((tunnel = g_hash_table_lookup(link->local_id_hash, &id)) == NULL)
        return NULL;

    return ref_get(tunnel);
}

void
tunnel_teardown(struct tunnel *tunnel, struct link *requestor)
{
    struct packet *pkt;

    if ( ( pkt = pkt_getref_alloc(0, 0)) == NULL)
        return;

    if ( tunnel->link_forward && requestor != tunnel->link_forward) {
        outgoing_link(pkt, tunnel->link_forward, LINK_TUNNEL_DESTROY,
                tunnel->remote_id_forward);
    }
    if ( tunnel->link_reverse && requestor != tunnel->link_reverse) {
        outgoing_link(pkt, tunnel->link_reverse, LINK_TUNNEL_DESTROY,
                tunnel->remote_id_reverse);
    }

    if ( tunnel->local_id_forward ) {
        g_hash_table_remove( tunnel->link_forward->local_id_hash, 
                                &tunnel->local_id_forward);

        if ( ! list_empty(&tunnel->fwd_iterator)  )
            list_del(&tunnel->fwd_iterator);

        ref_release(tunnel);
    }

    if ( tunnel->local_id_reverse ) {
        g_hash_table_remove( tunnel->link_reverse->local_id_hash, 
                                &tunnel->local_id_reverse);

        if ( ! list_empty(&tunnel->rev_iterator)  )
            list_del(&tunnel->rev_iterator);

        ref_release(tunnel);
    }

    if ( tunnel->state.have_fwd_init_id )
        g_hash_table_remove( tunnel->link_forward->neg_fwd_hash, 
                                 tunnel->forwarding_init_id);

    ref_release(pkt);
}

/*
 * XXX 
 *     this needs to be more careful.. it needs to work if its called
 *     more than once for a given tunnel.  (neg retransmissions)
 *
 * we always have to deal with sec.. either keep it for ourselves, and free it
 * at tunnel teardown, or free it now..
 */
static guint16
tunnel_finalize(struct tunnel *tunnel, struct secret *sec)
{
    struct keys *keys;

    keys = keys_setup(sec, CIPHER_BF, tunnel->state.neg_initiator,
            "I am a string, I am.");

    if ( keys == NULL ) {
        secret_free(sec); 
        return NEG_ERR_TEMP_FAILURE;
    } else 
        list_add_tail(&keys->multi_entry, &tunnel->multi_keys_head);

    l_debug("[ lf: 0x%08x, rf: 0x%08x, lr: 0x%08x, rr: 0x%08x]\n",
            tunnel->local_id_forward, tunnel->remote_id_forward,
            tunnel->local_id_reverse, tunnel->remote_id_reverse);

    if ( tunnel->local_id_forward && list_empty(&tunnel->fwd_iterator)) {

        g_hash_table_insert( tunnel->link_forward->local_id_hash, 
                &tunnel->local_id_forward, ref_get(tunnel));
        list_add(&tunnel->fwd_iterator, &tunnel->link_forward->tun_fwd_head);
    }
        
    if ( tunnel->local_id_reverse && list_empty(&tunnel->rev_iterator)) {

        g_hash_table_insert( tunnel->link_reverse->local_id_hash, 
                &tunnel->local_id_reverse, ref_get(tunnel));
        list_add(&tunnel->rev_iterator, &tunnel->link_reverse->tun_rev_head);
    }

    return NEG_ERR_OK;
}

/*
 * ------------------------------------------------------------
 * our interface into the negotiation code
 * ------------------------------------------------------------
 */
struct neg_callbacks tunnel_neg_cb;

static void
send_packet(struct packet *pkt, void *state)
{
    struct tunnel *tunnel = state;
    struct link *link = NULL;
    guint32 id;

    l_debug("sending neg as init %d tunneled %d\n",
            tunnel->state.neg_initiator, tunnel->state.tunneled_neg);

    if ( tunnel->state.neg_initiator ) {

        if ( tunnel->state.tunneled_neg )  {
            outgoing_tunnel_msgend(tunnel, tunnel->link_forward,
                    pkt, MSG_TYPE_NEG, tunnel->remote_id_forward);
            return;
        } 

        link = tunnel->link_forward;
        id = tunnel->local_id_forward;

    } else {
        id = tunnel->local_id_reverse;
        link = tunnel->link_reverse;

    }

    outgoing_link(pkt, link, LINK_TUNNEL_NEG, id);
}

static void
append_options(struct packet *pkt, void *state)
{
    struct tunnel *tunnel = state;
    guint32 next_hop = (guint32)g_slist_nth_data(tunnel->next_neighbour, 0);

    if ( next_hop )
        pkt_append_blob(pkt, B_NEXT_HOP, &next_hop, sizeof(next_hop));
}

/*
 * the server is receiving a confreq for it to look over.  
 * this is the last time the server sees it, so we have eto build up
 * IDs and addresses and things and assume that the client will like
 * what we send them.. they might send a destruction request otherwise,
 * or we'll just time the thing out.
 */
static guint16
parse_confreq(struct packet *pkt, struct packet *out,
        void *state, struct secret *sec)
{
    struct tunnel *tunnel = state;
    guint32 next_hop = 0;
    struct w_blob blob;

    if ( ! tunnel->local_id_reverse ) 
        tunnel->local_id_reverse = get_new_id(tunnel->link_reverse->local_id_hash);

    while ( pkt_get_head_blob(pkt, &blob) ) {
        switch (blob.type) {
            case B_NEXT_HOP:
                memcpy(&next_hop, pkt->head, sizeof(guint32));
                l_debug("was given next hop %x\n", next_hop);
                break;
            default: 
                break;
        }

        pkt_push_head(pkt, blob.len);
    }

    /*
     * XXX it seems bad things will happen if the confack sent from
     * here is lost and the client resends the confreq.. have to
     * think about that.
     */
    if ( next_hop ) {
        struct sockaddr_in sin;
        struct link *link;

        /*
         * we've already been bound to a next hop..
         */
        if ( tunnel->state.neg_forward ) {
            return NEG_ERR_NEXTHOP_EEXIST;
        }

        sin_init(&sin);
        sin.sin_addr.s_addr = next_hop;
        sin.sin_port = htons(CEBOLLA_PORT);

        link = link_nei_getref(TUNNEL_PARENT_SK(tunnel), &sin);
        if ( link == NULL )  {
            l_debug(" couldn'd find nei %s:%d\n", 
                    inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));
            return NEG_ERR_NEXTHOP_NO_LINK;
        }

        /*
         * we'll only forward further neg requests to the
         * next hop. 
         */
        tunnel->link_forward = link;
        tunnel->local_id_forward = get_new_id(tunnel->link_forward->local_id_hash);
        tunnel->state.neg_forward = 1;
    }

    /*
     * if we're the final hop we allocate an ip address
     */
    if ( next_hop == 0 ) {
        guint32 remote_addr;

        if ( ! tunnel->addr ) {
            struct socket *sk = TUNNEL_PARENT_SK(tunnel);
            if ((tunnel->addr = ip_alloc(sk->ipb)) == 0) {
                l_debug("couldn't allocate ip address\n");
                return NEG_ERR_TEMP_FAILURE;
            }
            ip_alloc(sk->ipb); /* XXX shadow alloc for remote side */
            tunnel->netmask = inet_addr("255.255.255.254"); /* XXX config */
        }

        remote_addr = htonl(ntohl(tunnel->addr) + 1);

        pkt_append_blob(out, B_IP_ADDR, &remote_addr, sizeof(guint32));
        pkt_append_blob(out, B_IP_NETMASK, &tunnel->netmask, sizeof(guint32));

        if ( tunnel->intf == NULL ) {
            if ( (tunnel->intf = int_allocate(tunnel)) == NULL )
                return NEG_ERR_TEMP_FAILURE;
        }
    }

    return tunnel_finalize(tunnel, sec);
}

/*
 * this completes a negotiation and brings a real tunnel
 * into being.
 */
static void
got_confack(struct packet *pkt, void *state, struct secret *sec)
{
    struct tunnel *tunnel = state;
    struct w_blob blob;
    guint32 next_hop;

    while ( pkt_get_head_blob(pkt, &blob) ) {
        switch (blob.type) {
            case B_IP_ADDR:
                memcpy(&tunnel->addr, pkt->head, sizeof(guint32));
                l_debug("was given addr %x\n", tunnel->addr);
                break;
            case B_IP_NETMASK:
                memcpy(&tunnel->netmask, pkt->head, sizeof(guint32));
                l_debug("was given netmask %x\n", tunnel->netmask);
                break;
            default: 
                break;
        }

        pkt_push_head(pkt, blob.len);
    }

    tunnel_finalize(tunnel, sec);

    next_hop = (guint32)g_slist_nth_data(tunnel->next_neighbour, 0);

    if ( next_hop ) {
        l_debug("initiating again..\n");

        tunnel->next_neighbour = g_slist_remove(tunnel->next_neighbour, 
                                                tunnel->next_neighbour->data);

        tunnel->state.tunneled_neg = 1;
        neg_initiate(&tunnel_neg_cb, tunnel);

        return;
    }

    if ( tunnel->intf == NULL ) {
        if ( (tunnel->intf = int_allocate(tunnel)) == NULL ) {
            l_debug("XXX failed to bring up interface..\n");
            return; /* XXX retry?  raise an error? */
        }
    }

}

static void
get_ref(void *state)
{
    struct tunnel *tunnel = state;
    ref_get(tunnel);
}

static void
release(void *state)
{
    struct tunnel *tunnel = state;
    ref_release(tunnel);
}

struct neg_callbacks tunnel_neg_cb = {
append_options:     append_options,
parse_confreq:      parse_confreq,
got_confack:        got_confack,
send_packet:        send_packet,
get_ref:            get_ref,
release:            release,
};

/*
 * this is called when we're the final hop in the path, but our 
 * negotiation specified an additional hop.  we're now receiving 
 * MSGEND tunnel messages with negotiations that we forward
 * on to the next hop over the link as normal tunnel negotiation
 */
void
tunnel_forward_neg(struct tunnel *tunnel, struct packet *pkt)
{
    struct w_neg_hdr nhdr;
    memcpy(&nhdr, pkt->head, sizeof(nhdr));

    /*
     * point to the tunnel with the init id so that we can find
     * the tunnel again as the respones come back over the
     * link.
     */

    if ( ! tunnel->state.have_fwd_init_id ) {
        memcpy(tunnel->forwarding_init_id, nhdr.initiator_id, NEG_ID_BYTES);

        g_hash_table_insert( tunnel->link_forward->neg_fwd_hash, 
                tunnel->forwarding_init_id, tunnel);

        hexdump(nhdr.initiator_id, NEG_ID_BYTES,
                "forwarding neg out link %p\n", tunnel->link_forward);

        tunnel->state.have_fwd_init_id  = 1;

    } else {
        /*
         * the initiator should only be forwarding one neg context through..
         * XXX that's probably not true if the first neg fails and they
         * retry.  crap.
         */
        if ( memcmp(tunnel->forwarding_init_id, 
                    nhdr.initiator_id, NEG_ID_BYTES) ) {
            l_debug("someone gave us a bad init id\n");
            return;
        }
    }

    outgoing_link(pkt, tunnel->link_forward, 
            LINK_TUNNEL_NEG, tunnel->local_id_forward);
}

void
tunnel_process_neg(struct link *link, struct packet *pkt, guint32 id)
{
    struct tunnel *tunnel;
    struct w_neg_hdr nhdr;

    memcpy(&nhdr, pkt->head, sizeof(nhdr));

    hexdump(pkt->head, NEG_ID_BYTES, "processing neg on link %p\n", link);

    /*
     * if the negotiation core has state then then the incoming neg packet is
     * in response to something we sent out.  the id announcement is for the
     * forward direction..
     */
    if ( (tunnel = neg_get_state_ref(pkt) ) ) {

        if ( ! tunnel->remote_id_forward )
            tunnel->remote_id_forward = id;

        neg_incoming(pkt, &tunnel_neg_cb, tunnel, &link, sizeof(link));
        ref_release(tunnel);

        return;
    }

    /*
     * otherwise, we might be seeing the response to a tunnel negotiation
     * that we forwarded as the end of the tunnel.. see if we should
     * forward the response back down a tunnel.
     */
    if ( (tunnel = g_hash_table_lookup(link->neg_fwd_hash, nhdr.initiator_id)) ) {

        if ( ! tunnel->remote_id_forward )
            tunnel->remote_id_forward = id;

        if ( nhdr.type == NEG_CONFACK ) 
            tunnel->state.saw_confack = 1;

        outgoing_tunnel_msgend(tunnel, tunnel->link_reverse, pkt, 
                MSG_TYPE_NEG, tunnel->remote_id_reverse);

        return;
    }
    
    /*
     * ok, we don't have state.  we're a responder and the id
     * advertisement is for the initiator's end.  we allocate
     * an initial tunnel to hold the announcement and to serve
     * as the proper tunnel when the negotiation is done.  this
     * is freed if it isn't needed.
     */
    if ( (tunnel = tunnel_getref_alloc(link, NULL))  == NULL) 
        return;

    if ( ! tunnel->remote_id_reverse )
        tunnel->remote_id_reverse = id;

    neg_incoming(pkt, &tunnel_neg_cb, tunnel, &link, sizeof(link));
    ref_release(tunnel);
}

int 
tunnel_create(struct link *link, int num_addrs, guint32 *addrs)
{
    struct tunnel *tunnel;
    int i;

    if ( (tunnel = tunnel_getref_alloc(NULL, link))  == NULL) 
        return -ENOMEM;

    l_debug("tun %p %d additional hops\n", tunnel, num_addrs);

    for ( i = 0 ; i < num_addrs ; i++ ) {
        struct sockaddr_in sin;
        sin.sin_addr.s_addr = addrs[i];
        l_debug("   %s\n", inet_ntoa(sin.sin_addr));

        tunnel->next_neighbour = g_slist_prepend(tunnel->next_neighbour,
                (gpointer)addrs[i]);
    }

    tunnel->local_id_forward = get_new_id(tunnel->link_forward->local_id_hash);

    tunnel->state.neg_initiator = 1;
    neg_initiate(&tunnel_neg_cb, tunnel);

    ref_release(tunnel);
    return 0;
}

int
tunnel_init(void)
{
    return 1;
}
