/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 *
 *    (c) Copyright 2002 Zach Brown <zab@zabbo.net>
 *    (c) Copyright 2001 Phil Schwan <phil@off.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <glib.h>

#include "log.h"
#include "main.h"
#include "pkt.h"
#include "incoming.h"
#include "link.h"
#include "udp.h"
#include "debug.h"
#include "topo.h"
#include "config.h"

#define MAX_UDP_LEN (64*1024)

static void
socket_destroy(struct socket *sk)
{
    if(sk->pfd.fd > -1)
        close(sk->pfd.fd);
    free(sk);
}

/*
 * some outgoing routines
 */

void
socket_sendto(struct socket *sk, struct sockaddr_in *sin, struct packet *pkt)
{
    hexdump(pkt->head, pkt->len, "============== %s:%d\n",
            inet_ntoa(sin->sin_addr), ntohs(sin->sin_port));

    if (sendto(sk->pfd.fd, pkt->head, pkt->len, 0, 
               (struct sockaddr *)sin, sizeof(struct sockaddr_in)) < 0)
        l_warn("sendto() failed: %s\n", g_strerror(errno));
}


/*
 * glib event loop goop
 */

static gboolean 
pfd_prepare(GSource *source, gint *timeout)
{
    *timeout = -1; /* block forever */
    return FALSE;  /* we have to go to poll to find out if we're ready */
}

/* FIXME: GSource is technically marked private, but how else am I supposed to
 * do this?  They just mean ``don't _change_ stuff directly,'' right? --phil */
static gboolean 
pfd_check(GSource *source)
{
    GSList *l;

    for (l = source->poll_fds; l != NULL; l = l->next) {
        GPollFD *pfd = (GPollFD *)l->data;

        if (pfd->revents != 0)
            return 1;
    }

    return 0;
}

static char error_msg[MAX_UDP_LEN];

static gboolean 
pfd_dispatch(GSource *source, GSourceFunc callback, gpointer user_data)
{
    struct socket *sk = (struct socket *)user_data;
    GPollFD *pfd = &sk->pfd;
    struct sockaddr_in from;
    socklen_t fromlen = sizeof(struct sockaddr_in);
    struct packet *pkt;
    int len;

    if (pfd->revents & G_IO_ERR) {
        struct msghdr msg;
        struct iovec iov;
        int len;

        msg.msg_name = NULL;
        msg.msg_namelen = 0;
        msg.msg_iov = &iov;
        msg.msg_iovlen = 1;
        iov.iov_base = error_msg;

        iov.iov_len = 8192;
        msg.msg_control = error_msg;
        msg.msg_controllen = 8192;

        while ( ( len = recvmsg(pfd->fd, &msg, MSG_ERRQUEUE)) > 0 ) {
            l_warn("received udp error\n");
        }
    }

    if ( (pkt = pkt_getref_alloc(0, MAX_UDP_LEN)) == NULL)
        return TRUE;

    sin_init(&from);

    if ( (len = recvfrom(pfd->fd, pkt->head, MAX_UDP_LEN, 0,
                         (struct sockaddr *)&from, &fromlen)) > 0) {

        pkt_push_tail(pkt, len);
        incoming_link(sk, pkt, &from);
    }

    ref_release(pkt);

    return TRUE;
}

static GSourceFuncs pfd_source_funcs = {
    prepare: pfd_prepare,
    check: pfd_check,
    dispatch: pfd_dispatch,
};

/*
 *  init calls done at startup..
 */
struct socket *
socket_prepare(struct daemon *dae, struct sockaddr_in *sin)
{
    struct socket *sk;

    if ((sk = g_new0(struct socket, 1)) == NULL) {
        l_warn("couldn't alloc socket data for %s:%d\n",
               inet_ntoa(sin->sin_addr), ntohs(sin->sin_port));
        return NULL;
    }

    sk->link_id_hash = g_hash_table_new_full(u32_hash, u32_equal,
                                        NULL, link_hash_delete);
    sk->link_nei_hash = g_hash_table_new_full(sin_hash, sin_equal,
                                        NULL, link_hash_delete);

    if ( sk->link_nei_hash == NULL || sk->link_id_hash == NULL )
        goto error;

    sk->pfd.fd = -1;
    memcpy(&sk->sin, sin, sizeof(struct sockaddr_in));
    INIT_LIST_HEAD(&sk->links_head);

    dae->socket_list = g_list_prepend(dae->socket_list, sk);

    return sk;

error:

    if ( sk->link_id_hash )
        g_hash_table_destroy(sk->link_id_hash); 
    if ( sk->link_nei_hash )
        g_hash_table_destroy(sk->link_nei_hash); 

    return NULL;
}

int
socket_init(struct daemon *dae)
{
    GSource *source;
    GList *list;

    /* One GSource per FD means glib manages the list instead of us */
    for (list = dae->socket_list; list != NULL; list = list->next) {
        struct socket *sk = (struct socket *)list->data;

        if ( topo_init(dae, sk) ) {
            l_warn("topo_init failed, dropping socket on the floor.\n");
            goto error;
        }

        if ((sk->pfd.fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
            l_warn("couldn't open datagram socket: %s\n", g_strerror(errno));
            goto error;
        }

        if (bind(sk->pfd.fd, (struct sockaddr *)&sk->sin, sizeof(sk->sin))) {
            l_warn("couldn't bind (%s:%d): %s\n", inet_ntoa(sk->sin.sin_addr),
                   ntohs(sk->sin.sin_port), g_strerror(errno));
            goto error;
        }

        l_debug("bound to udp %s:%d\n", inet_ntoa(sk->sin.sin_addr),
                ntohs(sk->sin.sin_port));

        sk->pfd.events = G_IO_IN | G_IO_HUP | G_IO_ERR;

        source = g_source_new(&pfd_source_funcs, sizeof(GSource));
        if (source == NULL) {
            l_warn("unable to allocate GSource\n");
            goto error;
        }

        g_source_add_poll(source, &sk->pfd);
        g_source_set_priority(source, G_PRIORITY_DEFAULT);
        g_source_set_callback(source, NULL, sk, NULL);
        g_source_attach(source, NULL);

        continue;

    error:
        list = list->prev;
        dae->socket_list = g_list_remove(dae->socket_list, sk);
        socket_destroy(sk);

        if (list == NULL)
            break; /* please don't segfault, mister. */
    }

    return 1; /* always returning success? */
}
